# This file is part of 'NTLM Authorization Proxy Server' http://sourceforge.net/projects/ntlmaps/
# Copyright 2001 Dmitry A. Rozmanov <dima@xenon.spb.ru>
#
# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.
try:
    from . import des_c
except ValueError:
    import des_c

#---------------------------------------------------------------------
class DES:

    des_c_obj = None

    #-----------------------------------------------------------------
    def __init__(self, key_str):
        ""
        k = str_to_key56(key_str)
        k = key56_to_key64(k)
        key_str = b''
        for i in k:
            key_str += bytes((i & 0xFF, ))
        self.des_c_obj = des_c.DES(key_str)

    #-----------------------------------------------------------------
    def encrypt(self, plain_text):
        ""
        return self.des_c_obj.encrypt(plain_text)

    #-----------------------------------------------------------------
    def decrypt(self, crypted_text):
        ""
        return self.des_c_obj.decrypt(crypted_text)

#---------------------------------------------------------------------
#Some Helpers
#---------------------------------------------------------------------

DESException = 'DESException'

#---------------------------------------------------------------------
def str_to_key56(key_str):
    ""
    if type(key_str) != type(''):
        #rise DESException, 'ERROR. Wrong key type.'
        pass
    if len(key_str) < 7:
        key_str = key_str + b'\000\000\000\000\000\000\000'[:(7 - len(key_str))]
    key_56 = []
    for i in key_str[:7]: key_56.append(i)

    return key_56

#---------------------------------------------------------------------
def key56_to_key64(key_56):
    ""
    key = []
    for i in range(8): key.append(0)

    key[0] = key_56[0];
    key[1] = ((key_56[0] << 7) & 0xFF) | (key_56[1] >> 1);
    key[2] = ((key_56[1] << 6) & 0xFF) | (key_56[2] >> 2);
    key[3] = ((key_56[2] << 5) & 0xFF) | (key_56[3] >> 3);
    key[4] = ((key_56[3] << 4) & 0xFF) | (key_56[4] >> 4);
    key[5] = ((key_56[4] << 3) & 0xFF) | (key_56[5] >> 5);
    key[6] = ((key_56[5] << 2) & 0xFF) | (key_56[6] >> 6);
    key[7] =  (key_56[6] << 1) & 0xFF;

    key = set_key_odd_parity(key)

    return key

#---------------------------------------------------------------------
def set_key_odd_parity(key):
    ""
    for i in range(len(key)):
        for k in range(7):
            bit = 0
            t = key[i] >> k
            bit = (t ^ bit) & 0x1
        key[i] = (key[i] & 0xFE) | bit

    return key



'''
See COPYRIGHT.md for copyright information.
'''
from arelle import (ViewWinTree, ModelObject)
from tkinter import TRUE

def viewProperties(modelXbrl, tabWin):
    modelXbrl.modelManager.showStatus(_("viewing properties"))
    view = ViewProperties(modelXbrl, tabWin)
    view.treeView["columns"] = ("value")
    view.treeView.column("#0", width=75, anchor="w")
    view.treeView.heading("#0", text="Property")
    view.treeView.column("value", width=600, anchor="w")
    view.treeView.heading("value", text="Value")
    view.treeView["displaycolumns"] = ("value")
    view.view()

class ViewProperties(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin):
        super(ViewProperties, self).__init__(modelXbrl, tabWin, "Properties", True)
        self.openProperties = set()

    def view(self):
        self.viewProperties(None, "")

    def cleanPreviousNodes(self,parentNode):
        for previousNode in self.treeView.get_children(parentNode):
            self.cleanPreviousNodes(previousNode)
            text = self.treeView.item(previousNode,'text')
            if str(self.treeView.item(previousNode,'open')) in ('true','1'): self.openProperties.add(text)
            else: self.openProperties.discard(text)
            self.treeView.delete(previousNode)

    def viewProperties(self, modelObject, parentNode):
        try:
            self.cleanPreviousNodes(parentNode)
        except Exception:
            pass    # possible tkinter issues
        if modelObject is not None and hasattr(modelObject, "propertyView"):
            self.showProperties(modelObject.propertyView, parentNode, 1)

    def showProperties(self, properties, parentNode, id):
        for tuple in properties:
            if tuple:
                lenTuple = len(tuple)
                if 2 <= lenTuple <= 3:
                    strId = str(id)
                    node = self.treeView.insert(parentNode, "end", strId, text=tuple[0], tags=("odd" if id & 1 else "even",))
                    self.treeView.set(strId, "value", tuple[1])
                    id += 1;
                    if lenTuple == 3:
                        if tuple[0] in self.openProperties:
                            self.treeView.item(node,open=True)
                        id = self.showProperties(tuple[2], node, id)
        return id

    def viewModelObject(self, modelObject):
        self.viewProperties(modelObject, "")



"""
See COPYRIGHT.md for copyright information.
"""
from tkinter import E, EW, N, S, Toplevel, W, messagebox

try:
    from tkinter.ttk import Frame, Button, Label, Entry
except ImportError:
    from ttk import Frame, Button, Label, Entry

import regex as re

from arelle import Locale
from arelle.CntlrWinTooltip import ToolTip
from arelle.UiUtil import checkbox, gridCombobox, gridHdr, label
from arelle.typing import TypeGetText

_: TypeGetText

'''
allow user to override system language codes for user interface and labels
'''


def askLanguage(mainWin):
    DialogLanguage(mainWin)


class DialogLanguage(Toplevel):
    def __init__(self, mainWin):
        super(DialogLanguage, self).__init__(mainWin.parent)
        self.mainWin = mainWin
        self.parent = mainWin.parent
        parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", self.parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.transient(self.parent)
        self.title(_("arelle - User Interface and Labels language code settings"))
        self.languageCodes = Locale.languageCodes()
        if systemLocales := Locale.availableLocales():  # unix/Mac locale -a supported locale codes
            localeOptionTitles = [
                title
                for title, labelCode in self.languageCodes.items()
                if Locale.bcp47LangToPosixLocale(labelCode) in systemLocales
            ]
        else:
            localeOptionTitles = self.languageCodes.keys()
        localeOptions = (_("System default locale ({0})").format(mainWin.modelManager.defaultLang), *sorted(localeOptionTitles))
        labelLanguageOptions = (_("System default language ({0})").format(mainWin.modelManager.defaultLang), *sorted(self.languageCodes.keys()))
        self.uiLang = Locale.posixLocaleToBCP47Lang(mainWin.config.get("userInterfaceLangOverride", ""))
        if self.uiLang == "" or self.uiLang == mainWin.modelManager.defaultLang:
            self.uiLangIndex = 0
        else:
            self.uiLangIndex = None
            for i, langName in enumerate(localeOptions):
                if i > 0 and self.uiLang == self.languageCodes.get(langName):
                    self.uiLangIndex = i
                    break
        self.labelLang = mainWin.config.get("labelLangOverride", "")
        if self.labelLang == "" or self.labelLang == mainWin.modelManager.defaultLang:
            self.labelLangIndex = 0
        else:
            self.labelLangIndex = None
            for i, langName in enumerate(labelLanguageOptions):
                if i > 0 and self.labelLang == self.languageCodes.get(langName):
                    self.labelLangIndex = i
                    break

        frame = Frame(self)

        defaultLanguage = mainWin.modelManager.defaultLang
        for langName, langCode in self.languageCodes.items():
            if mainWin.modelManager.defaultLang == langCode:
                defaultLanguage += ", " + langName
                break
        gridHdr(frame, 0, 0, _(
                 "The system default language is: {0} \n\n"
                 "You may override with a different language for user interface language and locale settings, and for language of taxonomy linkbase labels to display. \n\n").format(
                defaultLanguage),
              columnspan=5, wraplength=400)
        label(frame, 0, 1, _("User Interface:"))
        self.cbUiLang = gridCombobox(frame, 1, 1, values=localeOptions, selectindex=self.uiLangIndex, columnspan=4)
        label(frame, 0, 2, _("Labels:"))
        self.cbLabelLang = gridCombobox(frame, 1, 2, values=labelLanguageOptions, selectindex=self.labelLangIndex, columnspan=4)
        self.cbUiLang.focus_set()
        self.cbDisableRtl = checkbox(frame, 0, 3,  _('Disable rtl String'), 'disableRtlSting')
        ToolTip(self.cbDisableRtl, _('Disable reversing string read order for right to left languages, useful for some locale settings.'), wraplength=240)
        self.cbDisableRtl.valueVar.set(self.mainWin.config.get('disableRtl', 0))
        okButton = Button(frame, text=_("OK"), command=self.ok)
        cancelButton = Button(frame, text=_("Cancel"), command=self.close)
        okButton.grid(row=3, column=2, sticky=E, pady=3)
        cancelButton.grid(row=3, column=3, columnspan=2, sticky=EW, pady=3, padx=3)
        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(1, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def ok(self, event=None):
        self.mainWin.config['disableRtl'] = self.cbDisableRtl.value
        Locale.setDisableRTL(self.cbDisableRtl.value)
        labelLangIndex = self.cbLabelLang.valueIndex
        if labelLangIndex >= 0 and labelLangIndex != self.labelLangIndex: # changed
            if labelLangIndex == 0:
                langCode = self.mainWin.modelManager.defaultLang
            else:
                langCode = self.languageCodes[self.cbLabelLang.value]
            self.mainWin.config["labelLangOverride"] = langCode
            self.mainWin.labelLang = langCode

        uiLangIndex = self.cbUiLang.valueIndex
        if uiLangIndex >= 0 and uiLangIndex != self.uiLangIndex: # changed
            if uiLangIndex == 0:
                langCode = self.mainWin.modelManager.defaultLang
            else:
                langCode = self.languageCodes[self.cbUiLang.value]

            localeCode = Locale.findCompatibleLocale(langCode)
            if localeCode is not None:
                newLocale = Locale.getUserLocale(localeCode)
                self.mainWin.modelManager.locale = newLocale
            else:
                messagebox.showerror(_("User interface locale error"),
                                     _("Locale setting {0} is not supported on this system")
                                     .format(localeCode),
                                     parent=self)
                return
            if uiLangIndex != 0: # not the system default
                self.mainWin.config["userInterfaceLangOverride"] = localeCode
            else: # use system default
                self.mainWin.config.pop("userInterfaceLangOverride", None)
            self.mainWin.setUiLanguage(localeCode)

            if messagebox.askyesno(
                    _("User interface language changed"),
                    _("Should Arelle restart with changed user interface language, if there are any unsaved changes they would be lost!"),
                   parent=self):
                self.mainWin.uiThreadQueue.put((self.mainWin.quit, [None, True]))
            else:
                messagebox.showwarning(
                    _("User interface language changed"),
                    _("Please restart Arelle for the change in user interface language."),
                   parent=self)
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()



'''
See COPYRIGHT.md for copyright information.

based on pull request 4

'''
from __future__ import annotations

from importlib.metadata import EntryPoint
from tkinter import Toplevel, font, messagebox, VERTICAL, HORIZONTAL, N, S, E, W
from tkinter.constants import DISABLED, ACTIVE

from arelle.PluginManager import EntryPointRef

try:
    from tkinter.ttk import Treeview, Scrollbar, Frame, Label, Button
except ImportError:
    from ttk import Treeview, Scrollbar, Frame, Label, Button
from arelle import PluginManager, DialogURL, DialogOpenArchive
from arelle.CntlrWinTooltip import ToolTip
import os, time
import regex as re
EMPTYLIST = []
GROUPSEP = '\x01d'

def dialogPluginManager(mainWin):
    # check for updates in background
    import threading
    thread = threading.Thread(target=lambda cntlr=mainWin: backgroundCheckForUpdates(cntlr))
    thread.daemon = True
    thread.start()

def backgroundCheckForUpdates(cntlr):
    cntlr.showStatus(_("Checking for updates to plug-ins")) # clear web loading status
    modulesWithNewerFileDates = PluginManager.modulesWithNewerFileDates()
    if modulesWithNewerFileDates:
        cntlr.showStatus(_("Updates are available for these plug-ins: {0}")
                              .format(', '.join(modulesWithNewerFileDates)), clearAfter=5000)
    else:
        cntlr.showStatus(_("No updates found for plug-ins."), clearAfter=5000)
    time.sleep(0.1) # Mac locks up without this, may be needed for empty ui queue?
    cntlr.uiThreadQueue.put((DialogPluginManager, [cntlr, modulesWithNewerFileDates]))

class DialogPluginManager(Toplevel):
    def __init__(self, mainWin, modulesWithNewerFileDates):
        super(DialogPluginManager, self).__init__(mainWin.parent)

        self.ENABLE = _("Enable")
        self.DISABLE = _("Disable")
        self.parent = mainWin.parent
        self.cntlr = mainWin

        # copy plugins for temporary display
        self.pluginConfig = PluginManager.pluginConfig
        self.pluginConfigChanged = False
        self.uiClassMethodsChanged = False
        self.modelClassesChanged = False
        self.customTransformsChanged = False
        self.disclosureSystemTypesChanged = False
        self.hostSystemFeaturesChanged = False
        self.modulesWithNewerFileDates = modulesWithNewerFileDates

        parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", self.parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))

        self.title(_("Plug-in Manager"))
        frame = Frame(self)

        # left button frame
        buttonFrame = Frame(frame, width=40)
        buttonFrame.columnconfigure(0, weight=1)
        addLabel = Label(buttonFrame, text=_("Find plug-in modules:"), wraplength=60, justify="center")
        addSelectLocalButton = Button(buttonFrame, text=_("Select"), command=self.selectLocally)
        ToolTip(addSelectLocalButton, text=_("Select python module files from the local plugin directory."), wraplength=240)
        addBrowseLocalButton = Button(buttonFrame, text=_("Browse"), command=self.browseLocally)
        ToolTip(addBrowseLocalButton, text=_("File chooser allows browsing and selecting python module files to add (or reload) plug-ins, from the local file system."), wraplength=240)
        addWebButton = Button(buttonFrame, text=_("On Web"), command=self.findOnWeb)
        ToolTip(addWebButton, text=_("Dialog to enter URL full path to load (or reload) plug-ins, from the web or local file system."), wraplength=240)
        addLabel.grid(row=0, column=0, pady=4)
        addSelectLocalButton.grid(row=1, column=0, pady=4)
        addBrowseLocalButton.grid(row=2, column=0, pady=4)
        addWebButton.grid(row=3, column=0, pady=4)
        buttonFrame.grid(row=0, column=0, rowspan=3, sticky=(N, S, W), padx=3, pady=3)

        # right tree frame (plugins already known to arelle)
        modulesFrame = Frame(frame, width=720)
        vScrollbar = Scrollbar(modulesFrame, orient=VERTICAL)
        hScrollbar = Scrollbar(modulesFrame, orient=HORIZONTAL)
        self.modulesView = Treeview(modulesFrame, xscrollcommand=hScrollbar.set, yscrollcommand=vScrollbar.set, height=7)
        self.modulesView.grid(row=0, column=0, sticky=(N, S, E, W))
        self.modulesView.bind('<<TreeviewSelect>>', self.moduleSelect)
        hScrollbar["command"] = self.modulesView.xview
        hScrollbar.grid(row=1, column=0, sticky=(E,W))
        vScrollbar["command"] = self.modulesView.yview
        vScrollbar.grid(row=0, column=1, sticky=(N,S))
        modulesFrame.columnconfigure(0, weight=1)
        modulesFrame.rowconfigure(0, weight=1)
        modulesFrame.grid(row=0, column=1, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        self.modulesView.focus_set()

        self.modulesView.column("#0", width=120, anchor="w")
        self.modulesView.heading("#0", text=_("Name"))
        self.modulesView["columns"] = ("author", "ver", "status", "date", "update", "descr", "license")
        self.modulesView.column("author", width=100, anchor="w", stretch=False)
        self.modulesView.heading("author", text=_("Author"))
        self.modulesView.column("ver", width=60, anchor="w", stretch=False)
        self.modulesView.heading("ver", text=_("Version"))
        self.modulesView.column("status", width=50, anchor="w", stretch=False)
        self.modulesView.heading("status", text=_("Status"))
        self.modulesView.column("date", width=70, anchor="w", stretch=False)
        self.modulesView.heading("date", text=_("File Date"))
        self.modulesView.column("update", width=50, anchor="w", stretch=False)
        self.modulesView.heading("update", text=_("Update"))
        self.modulesView.column("descr", width=200, anchor="w", stretch=False)
        self.modulesView.heading("descr", text=_("Description"))
        self.modulesView.column("license", width=70, anchor="w", stretch=False)
        self.modulesView.heading("license", text=_("License"))

        classesFrame = Frame(frame)
        vScrollbar = Scrollbar(classesFrame, orient=VERTICAL)
        hScrollbar = Scrollbar(classesFrame, orient=HORIZONTAL)
        self.classesView = Treeview(classesFrame, xscrollcommand=hScrollbar.set, yscrollcommand=vScrollbar.set, height=5)
        self.classesView.grid(row=0, column=0, sticky=(N, S, E, W))
        hScrollbar["command"] = self.classesView.xview
        hScrollbar.grid(row=1, column=0, sticky=(E,W))
        vScrollbar["command"] = self.classesView.yview
        vScrollbar.grid(row=0, column=1, sticky=(N,S))
        classesFrame.columnconfigure(0, weight=1)
        classesFrame.rowconfigure(0, weight=1)
        classesFrame.grid(row=1, column=1, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        self.classesView.focus_set()

        self.classesView.column("#0", width=200, anchor="w")
        self.classesView.heading("#0", text=_("Class"))
        self.classesView["columns"] = ("modules",)
        self.classesView.column("modules", width=500, anchor="w", stretch=False)
        self.classesView.heading("modules", text=_("Modules"))

        # bottom frame module info details
        moduleInfoFrame = Frame(frame, width=700)
        moduleInfoFrame.columnconfigure(1, weight=1)

        self.moduleNameLabel = Label(moduleInfoFrame, wraplength=600, justify="left",
                                     font=font.Font(family='Helvetica', size=12, weight='bold'))
        self.moduleNameLabel.grid(row=0, column=0, columnspan=4, sticky=W)
        self.moduleAuthorHdr = Label(moduleInfoFrame, text=_("author:"), state=DISABLED)
        self.moduleAuthorHdr.grid(row=1, column=0, sticky=W)
        self.moduleAuthorLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleAuthorLabel.grid(row=1, column=1, columnspan=3, sticky=W)
        self.moduleDescrHdr = Label(moduleInfoFrame, text=_("description:"), state=DISABLED)
        self.moduleDescrHdr.grid(row=2, column=0, sticky=W)
        self.moduleDescrLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleDescrLabel.grid(row=2, column=1, columnspan=3, sticky=W)
        self.moduleClassesHdr = Label(moduleInfoFrame, text=_("classes:"), state=DISABLED)
        self.moduleClassesHdr.grid(row=3, column=0, sticky=W)
        self.moduleClassesLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleClassesLabel.grid(row=3, column=1, columnspan=3, sticky=W)
        ToolTip(self.moduleClassesLabel, text=_("List of classes that this plug-in handles."), wraplength=240)
        self.moduleVersionHdr = Label(moduleInfoFrame, text=_("version:"), state=DISABLED)
        self.moduleVersionHdr.grid(row=4, column=0, sticky=W)
        self.moduleVersionLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleVersionLabel.grid(row=4, column=1, columnspan=3, sticky=W)
        ToolTip(self.moduleVersionLabel, text=_("Version of plug-in module."), wraplength=240)
        self.moduleUrlHdr = Label(moduleInfoFrame, text=_("URL:"), state=DISABLED)
        self.moduleUrlHdr.grid(row=5, column=0, sticky=W)
        self.moduleUrlLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleUrlLabel.grid(row=5, column=1, columnspan=3, sticky=W)
        ToolTip(self.moduleUrlLabel, text=_("URL of plug-in module (local file path or web loaded file)."), wraplength=240)
        self.moduleDateHdr = Label(moduleInfoFrame, text=_("date:"), state=DISABLED)
        self.moduleDateHdr.grid(row=6, column=0, sticky=W)
        self.moduleDateLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleDateLabel.grid(row=6, column=1, columnspan=3, sticky=W)
        ToolTip(self.moduleDateLabel, text=_("Date of currently loaded module file (with parenthetical node when an update is available)."), wraplength=240)
        self.moduleLicenseHdr = Label(moduleInfoFrame, text=_("license:"), state=DISABLED)
        self.moduleLicenseHdr.grid(row=7, column=0, sticky=W)
        self.moduleLicenseLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleLicenseLabel.grid(row=7, column=1, columnspan=3, sticky=W)
        self.moduleImportsHdr = Label(moduleInfoFrame, text=_("imports:"), state=DISABLED)
        self.moduleImportsHdr.grid(row=8, column=0, sticky=W)
        self.moduleImportsLabel = Label(moduleInfoFrame, wraplength=600, justify="left")
        self.moduleImportsLabel.grid(row=8, column=1, columnspan=3, sticky=W)
        self.moduleEnableButton = Button(moduleInfoFrame, text=self.ENABLE, state=DISABLED, command=self.moduleEnable)
        ToolTip(self.moduleEnableButton, text=_("Enable/disable plug in."), wraplength=240)
        self.moduleEnableButton.grid(row=9, column=1, sticky=E)
        self.moduleReloadButton = Button(moduleInfoFrame, text=_("Reload"), state=DISABLED, command=self.moduleReload)
        ToolTip(self.moduleReloadButton, text=_("Reload/update plug in."), wraplength=240)
        self.moduleReloadButton.grid(row=9, column=2, sticky=E)
        self.moduleRemoveButton = Button(moduleInfoFrame, text=_("Remove"), state=DISABLED, command=self.moduleRemove)
        ToolTip(self.moduleRemoveButton, text=_("Remove plug in from plug in table (does not erase the plug in's file)."), wraplength=240)
        self.moduleRemoveButton.grid(row=9, column=3, sticky=E)
        moduleInfoFrame.grid(row=2, column=0, columnspan=5, sticky=(N, S, E, W), padx=3, pady=3)
        moduleInfoFrame.config(borderwidth=4, relief="groove")

        okButton = Button(frame, text=_("Close"), command=self.ok)
        ToolTip(okButton, text=_("Accept and changes (if any) and close dialog."), wraplength=240)
        cancelButton = Button(frame, text=_("Cancel"), command=self.close)
        ToolTip(cancelButton, text=_("Cancel changes (if any) and close dialog."), wraplength=240)
        okButton.grid(row=3, column=3, sticky=(S,E), pady=3)
        cancelButton.grid(row=3, column=4, sticky=(S,E), pady=3, padx=3)

        enableDisableFrame = Frame(frame)
        enableDisableFrame.grid(row=3, column=1, sticky=(S,W), pady=3)
        enableAllButton = Button(enableDisableFrame, text=_("Enable All"), command=self.enableAll)
        ToolTip(enableAllButton, text=_("Enable all plug ins."), wraplength=240)
        disableAllButton = Button(enableDisableFrame, text=_("Disable All"), command=self.disableAll)
        ToolTip(disableAllButton, text=_("Disable all plug ins."), wraplength=240)
        enableAllButton.grid(row=1, column=1)
        disableAllButton.grid(row=1, column=2)

        self.loadTreeViews()

        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))
        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(0, weight=0)
        frame.columnconfigure(1, weight=1)
        frame.rowconfigure(0, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        window.rowconfigure(0, weight=1)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def loadTreeViews(self):
        self.selectedModule = None

        # clear previous treeview entries
        for previousNode in self.modulesView.get_children(""):
            self.modulesView.delete(previousNode)

        def loadSubtree(parentNode, moduleItems):
            for moduleItem in sorted(moduleItems, key=lambda item: item[0]):
                moduleInfo = moduleItem[1]
                if parentNode or not moduleInfo.get("isImported"):
                    nodeName = moduleItem[0]
                    if parentNode:
                        nodeName = parentNode + GROUPSEP + nodeName
                    name = moduleInfo.get("name", nodeName)
                    node = self.modulesView.insert(parentNode, "end", nodeName, text=name)
                    self.modulesView.set(node, "author", moduleInfo.get("author"))
                    self.modulesView.set(node, "ver", moduleInfo.get("version"))
                    self.modulesView.set(node, "status", moduleInfo.get("status"))
                    self.modulesView.set(node, "date", moduleInfo.get("fileDate"))
                    if name in self.modulesWithNewerFileDates:
                        self.modulesView.set(node, "update", _("available"))
                    self.modulesView.set(node, "descr", moduleInfo.get("description"))
                    self.modulesView.set(node, "license", moduleInfo.get("license"))
                    if moduleInfo.get("imports"):
                        loadSubtree(node, [(importModuleInfo["name"],importModuleInfo)
                                           for importModuleInfo in moduleInfo["imports"]])

        loadSubtree("", self.pluginConfig.get("modules", {}).items())

        # clear previous treeview entries
        for previousNode in self.classesView.get_children(""):
            self.classesView.delete(previousNode)

        for i, classItem in enumerate(sorted(self.pluginConfig.get("classes", {}).items())):
            className, moduleList = classItem
            node = self.classesView.insert("", "end", className, text=className)
            self.classesView.set(node, "modules", ', '.join(moduleList))

        self.moduleSelect()  # clear out prior selection

    def ok(self, event=None):
        # check for orphaned classes (for which there is no longer a corresponding module)
        _moduleNames = self.pluginConfig.get("modules", {}).keys()
        _orphanedClassNames = set()
        for className, moduleList in self.pluginConfig.get("classes", {}).items():
            for _moduleName in moduleList.copy():
                if _moduleName not in _moduleNames: # it's orphaned
                    moduleList.remove(_moduleName)
                    self.pluginConfigChanged = True
            if not moduleList: # now orphaned
                _orphanedClassNames.add(className)
                self.pluginConfigChanged = True
        for _orphanedClassName in _orphanedClassNames:
            del self.pluginConfig["classes"][_orphanedClassName]

        if self.pluginConfigChanged:
            PluginManager.pluginConfig = self.pluginConfig
            PluginManager.pluginConfigChanged = True
            PluginManager.reset()  # force reloading of modules
        if self.uiClassMethodsChanged or self.modelClassesChanged or self.customTransformsChanged or self.disclosureSystemTypesChanged or self.hostSystemFeaturesChanged:  # may require reloading UI
            affectedItems = ""
            if self.uiClassMethodsChanged:
                affectedItems += _("menus of the user interface")
            if self.modelClassesChanged:
                if affectedItems:
                    affectedItems += _(" and ")
                affectedItems += _("model objects of the processor")
            if self.customTransformsChanged:
                if affectedItems:
                    affectedItems += _(" and ")
                affectedItems += _("custom transforms")
            if self.disclosureSystemTypesChanged:
                if affectedItems:
                    affectedItems += _(" and ")
                affectedItems += _("disclosure system types")
            if self.hostSystemFeaturesChanged:
                if affectedItems:
                    affectedItems += _(" and ")
                affectedItems += _("host system features")
            if messagebox.askyesno(_("User interface plug-in change"),
                                   _("A change in plug-in class methods may have affected {0}.  "
                                     "Please restart Arelle to due to these changes.  \n\n"
                                     "Should Arelle restart itself now "
                                     "(if there are any unsaved changes they would be lost!)?"
                                     ).format(affectedItems),
                                   parent=self):
                self.cntlr.uiThreadQueue.put((self.cntlr.quit, [None, True]))
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()

    def moduleSelect(self, *args):
        node = (self.modulesView.selection() or (None,))[0]
        if node:
            node = node.rpartition(GROUPSEP)[2] # drop leading path names for module name
        moduleInfo = self.pluginConfig.get("modules", {}).get(node)
        if moduleInfo:
            self.selectedModule = node
            name = moduleInfo["name"]
            self.moduleNameLabel.config(text=name)
            self.moduleAuthorHdr.config(state=ACTIVE)
            self.moduleAuthorLabel.config(text=moduleInfo.get("author"))
            self.moduleDescrHdr.config(state=ACTIVE)
            self.moduleDescrLabel.config(text=moduleInfo.get("description"))
            self.moduleClassesHdr.config(state=ACTIVE)
            self.moduleClassesLabel.config(text=', '.join(moduleInfo["classMethods"]))
            self.moduleVersionHdr.config(state=ACTIVE)
            self.moduleVersionLabel.config(text=moduleInfo.get("version"))
            self.moduleUrlHdr.config(state=ACTIVE)
            self.moduleUrlLabel.config(text=moduleInfo["moduleURL"])
            self.moduleDateHdr.config(state=ACTIVE)
            self.moduleDateLabel.config(text=moduleInfo["fileDate"] + " " +
                    (_("(an update is available)") if name in self.modulesWithNewerFileDates else ""))
            self.moduleLicenseHdr.config(state=ACTIVE)
            self.moduleLicenseLabel.config(text=moduleInfo.get("license"))
            if moduleInfo.get("imports"):
                self.moduleImportsHdr.config(state=ACTIVE)
                _text = ", ".join(mi["name"] for mi in moduleInfo["imports"][:3])
                if len(moduleInfo["imports"]) >= 3:
                    _text += ", ..."
                self.moduleImportsLabel.config(text=_text)
            _buttonState = DISABLED if moduleInfo.get("isImported") else ACTIVE
            self.moduleEnableButton.config(state=_buttonState,
                                           text={"enabled":self.DISABLE,
                                                 "disabled":self.ENABLE}[moduleInfo["status"]])
            self.moduleReloadButton.config(state=_buttonState)
            self.moduleRemoveButton.config(state=_buttonState)
        else:
            self.selectedModule = None
            self.moduleNameLabel.config(text="")
            self.moduleAuthorHdr.config(state=DISABLED)
            self.moduleAuthorLabel.config(text="")
            self.moduleDescrHdr.config(state=DISABLED)
            self.moduleDescrLabel.config(text="")
            self.moduleClassesHdr.config(state=DISABLED)
            self.moduleClassesLabel.config(text="")
            self.moduleVersionHdr.config(state=DISABLED)
            self.moduleVersionLabel.config(text="")
            self.moduleUrlHdr.config(state=DISABLED)
            self.moduleUrlLabel.config(text="")
            self.moduleDateHdr.config(state=DISABLED)
            self.moduleDateLabel.config(text="")
            self.moduleLicenseHdr.config(state=DISABLED)
            self.moduleLicenseLabel.config(text="")
            self.moduleImportsHdr.config(state=DISABLED)
            self.moduleImportsLabel.config(text="")
            self.moduleEnableButton.config(state=DISABLED, text=self.ENABLE)
            self.moduleReloadButton.config(state=DISABLED)
            self.moduleRemoveButton.config(state=DISABLED)

    @staticmethod
    def _choiceSortOrder(entryPointRef: EntryPointRef):
        moduleInfoMap = entryPointRef.moduleInfo
        key = moduleInfoMap["name"]
        group = {
            "ixbrl-viewer": "1",  # pip installed Arelle viewer
            "iXBRLViewerPlugin": "2",  # git clone installed Arelle viewer
            "EDGAR": "3",
            "Inline XBRL Document Set": "4",
            "XBRL rule processor (xule)": "5",
        }.get(key)
        if not group:
            if key.startswith("Validate"):
                group = "6"
            elif key.startswith("xbrlDB"):
                group = "7"
            else:
                group = "8"
        return group + key.lower()

    @staticmethod
    def _generateChoiceTuples(entryPointRefs: list[EntryPointRef]):
        """
        Generate list of choice tuples from list of entry point refs.
        :param entryPointRefs: List of entry point refs to convert to choice tuples.
        :return: List of choice tuples.
        """
        choiceTuples = []
        for entryPointRef in sorted(entryPointRefs, key=DialogPluginManager._choiceSortOrder):
            moduleInfo = entryPointRef.moduleInfo
            name = moduleInfo.get("name")
            path = moduleInfo.get("path")
            description = moduleInfo.get("description")
            version = moduleInfo.get("version")
            lic = moduleInfo.get("license")
            tooltip = "name: {}\ndescription: {}\nversion: {}\nlocation: {}\nlicense: {}".format(
                name, description, version, path, lic)
            choiceTuple = (name, tooltip, path, name, version, description, lic)
            choiceTuples.append(choiceTuple)
        return choiceTuples

    def selectLocally(self):
        entryPointRefs = EntryPointRef.discoverAll()
        choiceTuples = self._generateChoiceTuples(entryPointRefs)

        selectedPath = DialogOpenArchive.selectPlugin(self, choiceTuples)
        if selectedPath:
            if selectedPath.startswith(self.cntlr.pluginDir):
                selectedPath = selectedPath[len(self.cntlr.pluginDir)+1:]
            moduleInfo = PluginManager.moduleModuleInfo(moduleURL=selectedPath)
            self.loadFoundModuleInfo(moduleInfo, selectedPath)

    def browseLocally(self):
        initialdir = self.cntlr.pluginDir # default plugin directory
        if not self.cntlr.isMac: # can't navigate within app easily, always start in default directory
            initialdir = self.cntlr.config.setdefault("pluginOpenDir", initialdir)
        filename = self.cntlr.uiFileDialog("open",
                                           parent=self,
                                           title=_("Choose plug-in module file"),
                                           initialdir=initialdir,
                                           filetypes=[(_("Python files"), "*.py")],
                                           defaultextension=".py")
        if filename:
            # check if a package is selected (any file in a directory containing an __init__.py
            #if (os.path.basename(filename) == "__init__.py" and os.path.isdir(os.path.dirname(filename)) and
            #    os.path.isfile(filename)):
            #    filename = os.path.dirname(filename) # refer to the package instead
            self.cntlr.config["pluginOpenDir"] = os.path.dirname(filename)
            moduleInfo = PluginManager.moduleModuleInfo(moduleURL=filename)
            self.loadFoundModuleInfo(moduleInfo, filename)


    def findOnWeb(self):
        url = DialogURL.askURL(self)
        if url:  # url is the in-cache or local file
            moduleInfo = PluginManager.moduleModuleInfo(moduleURL=url)
            self.cntlr.showStatus("") # clear web loading status
            self.loadFoundModuleInfo(moduleInfo, url)

    def loadFoundModuleInfo(self, moduleInfo, url):
        if moduleInfo and moduleInfo.get("name"):
            self.addPluginConfigModuleInfo(moduleInfo)
            self.loadTreeViews()
        else:
            messagebox.showwarning(_("Module is not itself a plug-in or in a directory with package __init__.py plug-in.  "),
                                   _("File does not itself contain a python program with an appropriate __pluginInfo__ declaration: \n\n{0}")
                                   .format(url),
                                   parent=self)

    def checkIfImported(self, moduleInfo):
        if moduleInfo.get("isImported"):
            messagebox.showwarning(_("Plug-in is imported by a parent plug-in.  "),
                                   _("Plug-in has a parent, please request operation on the parent: \n\n{0}")
                                   .format(moduleInfo.get("name")),
                                   parent=self)
            return True
        return False

    def checkClassMethodsChanged(self, moduleInfo):
        for classMethod in moduleInfo["classMethods"]:
            if classMethod.startswith("CntlrWinMain.Menu"):
                self.uiClassMethodsChanged = True  # may require reloading UI
            elif classMethod == "ModelObjectFactory.ElementSubstitutionClasses":
                self.modelClassesChanged = True # model object factor classes changed
            elif classMethod == "ModelManager.LoadCustomTransforms":
                self.customTransformsChanged = True
            elif classMethod == "DisclosureSystem.Types":
                self.disclosureSystemTypesChanged = True # disclosure system types changed
            elif classMethod.startswith("Proxy."):
                self.hostSystemFeaturesChanged = True # system features (e.g., proxy) changed

    def removePluginConfigModuleInfo(self, name):
        moduleInfo = self.pluginConfig["modules"].get(name)
        if moduleInfo:
            if self.checkIfImported(moduleInfo):
                return;
            def _removePluginConfigModuleInfo(moduleInfo):
                _name = moduleInfo.get("name")
                if _name:
                    self.checkClassMethodsChanged(moduleInfo)
                    for classMethod in moduleInfo["classMethods"]:
                        classMethods = self.pluginConfig["classes"].get(classMethod)
                        if classMethods and _name in classMethods:
                            classMethods.remove(_name)
                            if not classMethods: # list has become unused
                                del self.pluginConfig["classes"][classMethod] # remove class
                    for importModuleInfo in moduleInfo.get("imports", EMPTYLIST):
                        _removePluginConfigModuleInfo(importModuleInfo)
                    self.pluginConfig["modules"].pop(_name, None)
            _removePluginConfigModuleInfo(moduleInfo)
            if not self.pluginConfig["modules"] and self.pluginConfig["classes"]:
                self.pluginConfig["classes"].clear() # clean orphan classes
            self.pluginConfigChanged = True

    def addPluginConfigModuleInfo(self, moduleInfo):
        if self.checkIfImported(moduleInfo):
            return;
        name = moduleInfo.get("name")
        self.removePluginConfigModuleInfo(name)  # remove any prior entry for this module
        def _addPlugin(moduleInfo):
            _name = moduleInfo.get("name")
            if _name:
                self.modulesWithNewerFileDates.discard(_name) # no longer has an update available
                self.pluginConfig["modules"][_name] = moduleInfo
                # add classes
                for classMethod in moduleInfo["classMethods"]:
                    classMethods = self.pluginConfig["classes"].setdefault(classMethod, [])
                    if name not in classMethods:
                        classMethods.append(_name)
                self.checkClassMethodsChanged(moduleInfo)
            for importModuleInfo in moduleInfo.get("imports", EMPTYLIST):
                _addPlugin(importModuleInfo)
        _addPlugin(moduleInfo)
        self.pluginConfigChanged = True

    def moduleEnable(self):
        if self.selectedModule in self.pluginConfig["modules"]:
            moduleInfo = self.pluginConfig["modules"][self.selectedModule]
            if self.checkIfImported(moduleInfo):
                return;
            def _moduleEnable(moduleInfo):
                if self.moduleEnableButton['text'] == self.ENABLE:
                    moduleInfo["status"] = "enabled"
                elif self.moduleEnableButton['text'] == self.DISABLE:
                    moduleInfo["status"] = "disabled"
                self.checkClassMethodsChanged(moduleInfo)
                for importModuleInfo in moduleInfo.get("imports", EMPTYLIST):
                    _moduleEnable(importModuleInfo) # set status on nested moduleInfo
                    if importModuleInfo['name'] in self.pluginConfig["modules"]: # set status on top level moduleInfo
                        _moduleEnable(self.pluginConfig["modules"][importModuleInfo['name']])
            _moduleEnable(moduleInfo)
            if self.moduleEnableButton['text'] == self.ENABLE:
                self.moduleEnableButton['text'] = self.DISABLE
            elif self.moduleEnableButton['text'] == self.DISABLE:
                self.moduleEnableButton['text'] = self.ENABLE
            self.pluginConfigChanged = True
            self.loadTreeViews()

    def moduleReload(self):
        if self.selectedModule in self.pluginConfig["modules"]:
            url = self.pluginConfig["modules"][self.selectedModule].get("moduleURL")
            if url:
                moduleInfo = PluginManager.moduleModuleInfo(moduleURL=url, reload=True)
                if moduleInfo:
                    if self.checkIfImported(moduleInfo):
                        return;
                    self.addPluginConfigModuleInfo(moduleInfo)
                    self.loadTreeViews()
                    self.cntlr.showStatus(_("{0} reloaded").format(moduleInfo["name"]), clearAfter=5000)
                else:
                    messagebox.showwarning(_("Module error"),
                                           _("File or module cannot be reloaded: \n\n{0}")
                                           .format(url),
                                           parent=self)

    def moduleRemove(self):
        if self.selectedModule in self.pluginConfig["modules"]:
            self.removePluginConfigModuleInfo(self.selectedModule)
            self.pluginConfigChanged = True
            self.loadTreeViews()

    def enableAll(self):
        self.enableDisableAll(True)

    def disableAll(self):
        self.enableDisableAll(False)

    def enableDisableAll(self, doEnable):
        for module in self.pluginConfig["modules"]:
            moduleInfo = self.pluginConfig["modules"][module]
            if not moduleInfo.get("isImported"):
                def _enableDisableAll(moduleInfo):
                    if doEnable:
                        moduleInfo["status"] = "enabled"
                    else:
                        moduleInfo["status"] = "disabled"
                    for importModuleInfo in moduleInfo.get("imports", EMPTYLIST):
                        _enableDisableAll(importModuleInfo)
                _enableDisableAll(moduleInfo)
                if doEnable:
                    self.moduleEnableButton['text'] = self.DISABLE
                else:
                    self.moduleEnableButton['text'] = self.ENABLE
        self.pluginConfigChanged = True
        self.loadTreeViews()



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import zipfile
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING

from arelle.Cntlr import Cntlr
from arelle.FileSource import FileSource
from arelle.ModelValue import QName
from arelle.typing import TypeGetText
from arelle.utils.PluginData import PluginData
from . import Constants
from .CoverItemRequirements import CoverItemRequirements
from .DeiRequirements import DeiRequirements
from .FilingFormat import FilingFormat
from .ReportFolderType import ReportFolderType
from .TableOfContentsBuilder import TableOfContentsBuilder
from .UploadContents import UploadContents, UploadPathInfo

if TYPE_CHECKING:
    from .ManifestInstance import ManifestInstance

_: TypeGetText


@dataclass
class ControllerPluginData(PluginData):
    _manifestInstancesById: dict[str, ManifestInstance]
    _tocBuilder: TableOfContentsBuilder
    _uploadContents: UploadContents | None
    _usedFilepaths: set[Path]

    def __init__(self, name: str):
        super().__init__(name)
        self._manifestInstancesById = {}
        self._tocBuilder = TableOfContentsBuilder()
        self._usedFilepaths = set()
        self._uploadContents = None

    def __hash__(self) -> int:
        return id(self)

    def addManifestInstance(self, manifestInstance: ManifestInstance) -> None:
        """
        Add a manifest instance with unique ID to the plugin data.
        """
        self._manifestInstancesById[manifestInstance.id] = manifestInstance

    @lru_cache(1)
    def getCoverItemRequirements(self, jsonPath: Path) -> CoverItemRequirements:
        return CoverItemRequirements(jsonPath)

    @lru_cache(1)
    def getDeiRequirements(self, csvPath: Path, deiItems: tuple[QName, ...], filingFormats: tuple[FilingFormat, ...]) -> DeiRequirements:
        return DeiRequirements(csvPath, deiItems, filingFormats)

    def getManifestInstances(self) -> list[ManifestInstance]:
        """
        Retrieve all loaded manifest instances.
        """
        return list(self._manifestInstancesById.values())

    def getTableOfContentsBuilder(self) -> TableOfContentsBuilder:
        return self._tocBuilder

    def getUploadContents(self) -> UploadContents | None:
        return self._uploadContents

    def setUploadContents(self, fileSource: FileSource) -> UploadContents:
        reports = defaultdict(list)
        uploadPaths = {}
        for path, zipPath in self.getUploadFilepaths(fileSource).items():
            if len(path.parts) == 0:
                continue
            assert isinstance(fileSource.basefile, str)
            fullPath = Path(fileSource.basefile) / path
            parents = list(reversed([p.name for p in path.parents if len(p.name) > 0]))
            reportFolderType = None
            isCorrection = True
            isDirectory = zipPath.is_dir()
            isInSubdirectory = False
            reportPath = None
            if len(parents) > 0:
                isCorrection = parents[0] != 'XBRL'
                if not isCorrection:
                    if len(parents) > 1:
                        formName = parents[1]
                        isInSubdirectory = len(parents) > 2
                        reportFolderType = ReportFolderType.parse(formName)
                if reportFolderType is None:
                    formName = parents[0]
                    isInSubdirectory = len(parents) > 1
                    reportFolderType = ReportFolderType.parse(formName)
                if reportFolderType is not None:
                    reportPath = Path(reportFolderType.value) if isCorrection else Path("XBRL") / reportFolderType.value
                    if not isCorrection:
                        reports[reportFolderType].append(path)
            uploadPaths[path] = UploadPathInfo(
                fullPath=fullPath,
                isAttachment=reportFolderType is not None and reportFolderType.isAttachment,
                isCorrection=isCorrection,
                isCoverPage=not isDirectory and path.stem.startswith(Constants.COVER_PAGE_FILENAME_PREFIX),
                isDirectory=isDirectory,
                isRoot=len(path.parts) == 1,
                isSubdirectory=isInSubdirectory or (isDirectory and reportFolderType is not None),
                path=path,
                reportFolderType=reportFolderType,
                reportPath=reportPath,
            )
        self._uploadContents = UploadContents(
            reports={k: frozenset(v) for k, v in reports.items() if len(v) > 0},
            uploadPaths=list(uploadPaths.values())
        )
        return self._uploadContents

    @lru_cache(1)
    def getUploadFilepaths(self, fileSource: FileSource) -> dict[Path, zipfile.Path]:
        if not self.isUpload(fileSource):
            return {}
        paths = {}
        assert isinstance(fileSource.fs, zipfile.ZipFile)
        # First, fill in paths from zip file list
        for file in fileSource.fs.filelist:
            zipPath = zipfile.Path(fileSource.fs, file.filename)
            paths[Path(file.filename)] = zipPath
        # Then, fill in any parent directories that weren't in file list
        for path in list(paths):
            for parent in path.parents:
                if parent in paths:
                    continue
                paths[parent] = zipfile.Path(fileSource.fs, parent.as_posix() + '/')
        return {
            path: paths[path]
            for path in sorted(paths)
        }

    @lru_cache(1)
    def getUploadFileSizes(self, fileSource: FileSource) -> dict[Path, int]:
        """
        Get the sizes of files in the upload directory.
        :param fileSource: The FileSource instance to get file sizes for.
        :return: A dictionary mapping file paths to their sizes.
        """
        if not self.isUpload(fileSource):
            return {}
        assert isinstance(fileSource.fs, zipfile.ZipFile)
        return {
            Path(i.filename): i.file_size
            for i in fileSource.fs.infolist()
            if not i.is_dir()
        }

    def getUsedFilepaths(self) -> frozenset[Path]:
        return frozenset(self._usedFilepaths)

    @lru_cache(1)
    def isUpload(self, fileSource: FileSource) -> bool:
        fileSource.open()  # Make sure file source is open
        if (fileSource.fs is None or
                not isinstance(fileSource.fs, zipfile.ZipFile)):
            if fileSource.cntlr is not None:
                fileSource.cntlr.addToLog(
                    _("The target file is not a zip file, so upload validation was not performed."),
                    messageCode="EDINET.uploadNotValidated",
                    file=str(fileSource.url)
                )
            return False
        return True

    def matchManifestInstance(self, ixdsDocUrls: list[str]) -> ManifestInstance | None:
        """
        Match a manifest instance based on the provided ixdsDocUrls.
        A one-to-one mapping must exist between the model's IXDS document URLs and the manifest instance's IXBRL files.
        :param ixdsDocUrls: A model's list of IXDS document URLs.
        :return: A matching ManifestInstance if found, otherwise None.
        """
        modelUrls = set(ixdsDocUrls)
        matchedInstance = None
        for instance in self._manifestInstancesById.values():
            if len(instance.ixbrlFiles) != len(ixdsDocUrls):
                continue
            manifestUrls = {str(path) for path in instance.ixbrlFiles}
            unmatchedModelUrls = set(modelUrls)
            unmatchedManifestUrls = set(manifestUrls)
            for modelUrl in modelUrls:
                if modelUrl not in unmatchedModelUrls:
                    continue
                for manifestUrl in manifestUrls:
                    if modelUrl.endswith(manifestUrl):
                        unmatchedModelUrls.remove(modelUrl)
                        unmatchedManifestUrls.remove(manifestUrl)
                        break
            if len(unmatchedModelUrls) > 0:
                continue
            if len(unmatchedManifestUrls) > 0:
                continue
            matchedInstance = instance
            break
        return matchedInstance

    def addUsedFilepath(self, path: Path) -> None:
        self._usedFilepaths.add(path)

    @staticmethod
    def get(cntlr: Cntlr, name: str) -> ControllerPluginData:
        controllerPluginData = cntlr.getPluginData(name)
        if controllerPluginData is None:
            controllerPluginData = ControllerPluginData(name)
            cntlr.setPluginData(controllerPluginData)
        assert isinstance(controllerPluginData, ControllerPluginData), "Expected ControllerPluginData instance."
        return controllerPluginData



"""
See COPYRIGHT.md for copyright information.
"""

from __future__ import annotations

from enum import Enum

from arelle.packages.PackageConst import META_INF_DIRECTORY

INLINE_XBRL_REPORT_PACKAGE_DOCUMENT_TYPE = "https://xbrl.org/report-package/2023/xbri"
NON_INLINE_XBRL_REPORT_PACKAGE_DOCUMENT_TYPE = "https://xbrl.org/report-package/2023/xbr"
UNCONSTRAINED_REPORT_PACKAGE_DOCUMENT_TYPE = "https://xbrl.org/report-package/2023"

SUPPORTED_DOCUMENT_TYPES = frozenset(
    [
        INLINE_XBRL_REPORT_PACKAGE_DOCUMENT_TYPE,
        NON_INLINE_XBRL_REPORT_PACKAGE_DOCUMENT_TYPE,
        UNCONSTRAINED_REPORT_PACKAGE_DOCUMENT_TYPE,
    ]
)

INLINE_XBRL_REPORT_PACKAGE_EXTENSION = ".xbri"
NON_INLINE_XBRL_REPORT_PACKAGE_EXTENSION = ".xbr"
UNCONSTRAINED_REPORT_PACKAGE_EXTENSION = ".zip"


REPORT_PACKAGE_EXTENSIONS = frozenset(
    [
        INLINE_XBRL_REPORT_PACKAGE_EXTENSION,
        NON_INLINE_XBRL_REPORT_PACKAGE_EXTENSION,
        UNCONSTRAINED_REPORT_PACKAGE_EXTENSION,
        UNCONSTRAINED_REPORT_PACKAGE_EXTENSION.upper(),
    ]
)

REPORT_PACKAGE_FILE = f"{META_INF_DIRECTORY}/reportPackage.json"
REPORTS_DIRECTORY = "reports"

XBRL_2_1_REPORT_FILE_EXTENSION = ".xbrl"
JSON_REPORT_FILE_EXTENSION = ".json"
NON_INLINE_REPORT_FILE_EXTENSIONS = frozenset([
    XBRL_2_1_REPORT_FILE_EXTENSION,
    JSON_REPORT_FILE_EXTENSION,
])
INLINE_REPORT_FILE_EXTENSIONS = frozenset([".xhtml", ".html", ".htm"])
REPORT_FILE_EXTENSIONS = frozenset(
    [
        *NON_INLINE_REPORT_FILE_EXTENSIONS,
        *INLINE_REPORT_FILE_EXTENSIONS,
    ]
)


class ReportType(Enum):
    INLINE_XBRL_REPORT_PACKAGE = INLINE_XBRL_REPORT_PACKAGE_EXTENSION
    NON_INLINE_XBRL_REPORT_PACKAGE = NON_INLINE_XBRL_REPORT_PACKAGE_EXTENSION
    UNCONSTRAINED_REPORT_PACKAGE = UNCONSTRAINED_REPORT_PACKAGE_EXTENSION

    @staticmethod
    def fromExtension(extension: str) -> ReportType | None:
        for reportType in ReportType:
            if reportType.value == extension.lower():
                return reportType
        return None

    @property
    def reportFileExtensions(self) -> frozenset[str]:
        if self == ReportType.INLINE_XBRL_REPORT_PACKAGE:
            return INLINE_REPORT_FILE_EXTENSIONS
        if self == ReportType.NON_INLINE_XBRL_REPORT_PACKAGE:
            return NON_INLINE_REPORT_FILE_EXTENSIONS
        if self == ReportType.UNCONSTRAINED_REPORT_PACKAGE:
            return REPORT_FILE_EXTENSIONS
        raise ValueError(f"Report type without defined report file extensions: {self}")

    @property
    def isConstrained(self) -> bool:
        return self != ReportType.UNCONSTRAINED_REPORT_PACKAGE



'''
See COPYRIGHT.md for copyright information.
'''
from arelle import ViewWinTree, XbrlConst
from arelle.ModelDtsObject import ModelRelationship
from arelle.ModelInstanceObject import ModelFact
from collections import defaultdict

def viewConcepts(modelXbrl, tabWin, header, lang=None, altTabWin=None):
    modelXbrl.modelManager.showStatus(_("viewing concepts"))
    view = ViewConcepts(modelXbrl, tabWin, header, lang)
    view.treeView["columns"] = ("conceptname", "id", "abstr", "subsGrp", "type", "periodType", "balance", "facets")
    view.treeView.column("#0", width=250, anchor="w")
    view.treeView.heading("#0", text="Label")
    view.treeView.column("conceptname", width=250, anchor="w", stretch=False)
    view.treeView.heading("conceptname", text="Name")
    view.treeView.column("id", width=250, anchor="w", stretch=False)
    view.treeView.heading("id", text="ID")
    view.treeView.column("abstr", width=50, anchor="center", stretch=False)
    view.treeView.heading("abstr", text="Abstract")
    view.treeView.column("subsGrp", width=290, anchor="w", stretch=False)
    view.treeView.heading("subsGrp", text="Subs Grp")
    view.treeView.column("type", width=200, anchor="w", stretch=False)
    view.treeView.heading("type", text="Type")
    view.treeView.column("periodType", width=80, anchor="w", stretch=False)
    view.treeView.heading("periodType", text="Period Type")
    view.treeView.column("balance", width=70, anchor="w", stretch=False)
    view.treeView.heading("balance", text="Balance")
    view.treeView.column("facets", width=200, anchor="w", stretch=False)
    view.treeView.heading("facets", text="Facets")
    view.treeView["displaycolumns"] = ("conceptname", "id", "abstr", "subsGrp", "type", "periodType", "balance", "facets")
    view.view()
    view.blockSelectEvent = 1
    view.blockViewModelObject = 0
    view.treeView.bind("<<TreeviewSelect>>", view.treeviewSelect, '+')
    view.treeView.bind("<Enter>", view.treeviewEnter, '+')
    view.treeView.bind("<Leave>", view.treeviewLeave, '+')

    # languages menu
    menu = view.contextMenu()
    view.menuAddClipboard()
    view.menuAddLangs()
    view.menuAddLabelRoles()
    view.menuAddNameStyle()
    view.menuAddViews(addClose=False, tabWin=altTabWin)

class ViewConcepts(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin, header, lang):
        super(ViewConcepts, self).__init__(modelXbrl, tabWin, header, True, lang)

    def view(self):
        # sort by labels
        self.setColumnsSortable()
        lbls = defaultdict(list)
        role = self.labelrole
        lang = self.lang
        nameIsPrefixed = self.nameIsPrefixed
        for concept in set(self.modelXbrl.qnameConcepts.values()): # may be twice if unqualified, with and without namespace
            lbls[concept.label(role,lang=lang)].append(concept.objectId())
        srtLbls = sorted(lbls.keys())
        '''
        self.nodeToObjectId = {}
        self.objectIdToNode = {}
        '''
        self.clearTreeView()
        nodeNum = 1
        excludedNamespaces = XbrlConst.ixbrlAll.union(
            (XbrlConst.xbrli, XbrlConst.link, XbrlConst.xlink, XbrlConst.xl,
             XbrlConst.xbrldt,
             XbrlConst.xhtml))
        for label in srtLbls:
            for objectId in lbls[label]:
                concept = self.modelXbrl.modelObject(objectId)
                if concept.modelDocument.targetNamespace not in excludedNamespaces:
                    '''
                    node = "node{0}".format(nodeNum)
                    objectId = concept.objectId()
                    label = concept.label(lang=self.lang)
                    self.nodeToObjectId[node] = objectId
                    self.objectIdToNode[objectId] = node
                    if self.treeView.exists(node):
                        self.treeView.item(node, text=label)
                    else:
                        node = self.treeView.insert("", "end", node, text=label)
                    '''
                    node = self.treeView.insert("", "end",
                                                concept.objectId(),
                                                text=concept.label(role,lang=lang,linkroleHint=XbrlConst.defaultLinkRole),
                                                tags=("odd" if nodeNum & 1 else "even",))
                    nodeNum += 1
                    self.treeView.set(node, "conceptname", concept.qname if nameIsPrefixed else concept.name)
                    self.treeView.set(node, "id", concept.id)
                    self.treeView.set(node, "abstr", concept.abstract)
                    self.treeView.set(node, "subsGrp", concept.substitutionGroupQname)
                    self.treeView.set(node, "type", concept.typeQname)
                    if concept.periodType:
                        self.treeView.set(node, "periodType", concept.periodType)
                    if concept.balance:
                        self.treeView.set(node, "balance", concept.balance)
                    facets = concept.facets
                    if facets:
                        self.treeView.set(node, "facets",
                            "\n".join("{0}={1}".format(
                                   name,
                                   sorted(value.keys()) if isinstance(value,dict) else value
                                   ) for name,value in sorted(facets.items()))
                            )

    def treeviewEnter(self, *args):
        self.blockSelectEvent = 0

    def treeviewLeave(self, *args):
        self.blockSelectEvent = 1

    def treeviewSelect(self, event):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            #self.modelXbrl.viewModelObject(self.nodeToObjectId[self.treeView.selection()[0]])
            self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            try:
                if isinstance(modelObject, ModelRelationship):
                    conceptId = modelObject.toModelObject.objectId()
                elif isinstance(modelObject, ModelFact):
                    conceptId = self.modelXbrl.qnameConcepts[modelObject.qname].objectId()
                else:
                    conceptId = modelObject.objectId()
                #node = self.objectIdToNode[conceptId]
                node = conceptId
                if self.treeView.exists(node):
                    self.treeView.see(node)
                    self.treeView.selection_set(node)
            except (AttributeError, KeyError):
                    self.treeView.selection_set(())
            self.blockViewModelObject -= 1



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Optional, Any, Union
from collections import OrderedDict
from arelle.ModelValue import QName, AnyURI
from arelle.PythonUtil import OrderedSet
from .XbrlTypes import XbrlDtsType, QNameKeyType
from .XbrlTaxonomyObject import XbrlReportObject

class XbrlFact(XbrlReportObject):
    report: XbrlDtsType
    id: Optional[str] # synthesized id (from fact key in JSON), marked Optional because it's a key, not value, in json source.
    value: Optional[str] # (required) The value of the {value} property of the fact. The value MUST be represented as (xbrlje:invalidFactValue):
    decimals: Optional[int] # An integer providing the value of the {decimals} property, or absent if the value is infinitely precise or not applicable (for nil or non-numeric facts).
    dimensions: dict[Union[QName, str], str] # (required) A dimensions object with properties corresponding to the members of the {dimensions} property.
    links: dict[str, dict[str, list[str]]] # A links object, cormesponding to the link groups associated with the fact. This member MAY be absent if there are no link groups associated with the fact.


def addReportProperties(dts: XbrlDtsType):
    dts.linkTypes: dict[str, AnyURI] = OrderedDict()
    dts.linkGroups: dict[str, AnyURI] = OrderedDict()
    dts.facts: dict[str, XbrlFact] = OrderedDict()



'''
See COPYRIGHT.md for copyright information.
'''
from __future__ import annotations

import ast
import gettext
from glob import glob
import importlib.util
import json
import logging
import os
import sys
import time
import traceback
import types
from collections import defaultdict
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from importlib.metadata import EntryPoint, entry_points
from numbers import Number
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, cast

import arelle.FileSource
from arelle.Locale import getLanguageCodes
from arelle.PythonUtil import isLegacyAbs
from arelle.typing import TypeGetText
from arelle.UrlUtil import isAbsolute

if TYPE_CHECKING:
    # Prevent potential circular import error
    from .Cntlr import Cntlr


_: TypeGetText

PLUGIN_TRACE_FILE = None
# PLUGIN_TRACE_FILE = "c:/temp/pluginerr.txt"
PLUGIN_TRACE_LEVEL = logging.WARNING

# plugin control is static to correspond to statically loaded modules
pluginJsonFile = None
pluginConfig: dict | None = None
pluginConfigChanged = False
pluginTraceFileLogger = None
modulePluginInfos = {}
pluginMethodsForClasses = {}
_cntlr = None
_pluginBase = None
EMPTYLIST = []
_ERROR_MESSAGE_IMPORT_TEMPLATE = "Unable to load module {}"

def init(cntlr: Cntlr, loadPluginConfig: bool = True) -> None:
    global pluginJsonFile, pluginConfig, pluginTraceFileLogger, modulePluginInfos, pluginMethodsForClasses, pluginConfigChanged, _cntlr, _pluginBase
    if PLUGIN_TRACE_FILE:
        pluginTraceFileLogger = logging.getLogger(__name__)
        pluginTraceFileLogger.propagate = False
        handler = logging.FileHandler(PLUGIN_TRACE_FILE)
        formatter = logging.Formatter('%(asctime)s.%(msecs)03dz [%(levelname)s] %(message)s', datefmt='%Y-%m-%dT%H:%M:%S')
        handler.setFormatter(formatter)
        handler.setLevel(PLUGIN_TRACE_LEVEL)
        pluginTraceFileLogger.addHandler(handler)
    pluginConfigChanged = False
    _cntlr = cntlr
    _pluginBase = cntlr.pluginDir + os.sep
    if loadPluginConfig:
        try:
            pluginJsonFile = cntlr.userAppDir + os.sep + "plugins.json"
            with open(pluginJsonFile, encoding='utf-8') as f:
                pluginConfig = json.load(f)
            freshenModuleInfos()
        except Exception:
            pass # on GAE no userAppDir, will always come here
    if not pluginConfig:
        pluginConfig = {  # savable/reloadable plug in configuration
            "modules": {}, # dict of moduleInfos by module name
            "classes": {}  # dict by class name of list of class modules in execution order
        }
        pluginConfigChanged = False # don't save until something is added to pluginConfig
    modulePluginInfos = {}  # dict of loaded module pluginInfo objects by module names
    pluginMethodsForClasses = {} # dict by class of list of ordered callable function objects

def reset() -> None:  # force reloading modules and plugin infos
    if modulePluginInfos:
        modulePluginInfos.clear()  # dict of loaded module pluginInfo objects by module names
    if pluginMethodsForClasses:
        pluginMethodsForClasses.clear() # dict by class of list of ordered callable function objects

def orderedPluginConfig():
    fieldOrder = [
        'name',
        'status',
        'fileDate',
        'version',
        'description',
        'moduleURL',
        'localeURL',
        'localeDomain',
        'license',
        'author',
        'copyright',
        'classMethods',
    ]
    priorityIndex = {k: i for i, k in enumerate(fieldOrder)}

    def sortModuleInfo(moduleInfo):
        # Prioritize known fields by the index in fieldOrder; sort others alphabetically
        orderedKeys = sorted(
            moduleInfo.keys(),
            key=lambda k: (priorityIndex.get(k, len(priorityIndex)), k)
        )
        return {k: moduleInfo[k] for k in orderedKeys}

    orderedModules = {
        moduleName: sortModuleInfo(pluginConfig['modules'][moduleName])
        for moduleName in sorted(pluginConfig['modules'].keys())
    }

    return {
        'modules': orderedModules,
        'classes': dict(sorted(pluginConfig['classes'].items()))
    }

def save(cntlr: Cntlr) -> None:
    global pluginConfigChanged
    if pluginConfigChanged and cntlr.hasFileSystem and not cntlr.disablePersistentConfig:
        pluginJsonFile = cntlr.userAppDir + os.sep + "plugins.json"
        with open(pluginJsonFile, 'w', encoding='utf-8') as f:
            jsonStr = str(json.dumps(orderedPluginConfig(), ensure_ascii=False, indent=2)) # might not be unicode in 2.7
            f.write(jsonStr)
        pluginConfigChanged = False

def close() -> None:  # close all loaded methods
    if pluginConfig is not None:
        pluginConfig.clear()
    if modulePluginInfos is not None:
        modulePluginInfos.clear()
    if pluginMethodsForClasses is not None:
        pluginMethodsForClasses.clear()
    global webCache
    webCache = None

''' pluginInfo structure:

__pluginInfo__ = {
    'name': (required)
    'version': (required)
    'description': (optional)
    'moduleURL': (required) # added by plug in manager, not in source file
    'localeURL': (optional) # L10N internationalization for this module (subdirectory if relative)
    'localeDomain': (optional) # domain for L10N internationalization (e.g., 'arelle')
    'license': (optional)
    'author': (optional)
    'copyright': (optional)
    # classes of mount points (required)
    'a.b.c': method (function) to do something
    'a.b.c.d' : method (function) to do something
    # import (plugins to be loaded for this package) (may be multiple imports like in python)
    'import': [string, list or tuple of URLs or relative file names of imported plug-ins]
}

moduleInfo = {
    'name': (required)
    'status': enabled | disabled
    'version': (required)
    'fileDate': 2000-01-01
    'description': (optional)
    'moduleURL': (required) # same as file path, can be a URL (of a non-package .py file or a package directory)
    'localeURL': (optional) # for L10N internationalization within module
    'localeDomain': (optional) # domain for L10N internationalization
    'license': (optional)
    'author': (optional)
    'copyright': (optional)
    'classMethods': [list of class names that have methods in module]
    'imports': [list of imported plug-in moduleInfos]
    'isImported': True if module was imported by a parent plug-in module
}


'''

def logPluginTrace(message: str, level: Number) -> None:
    """
    If plugin trace file logging is configured, logs `message` to it.
    Only logs to controller logger if log is an error.
    :param message: Message to be logged
    :param level: Log level of message (e.g. logging.INFO)
    """
    if pluginTraceFileLogger:
        pluginTraceFileLogger.log(level, message)
    if level >= logging.ERROR:
        _cntlr.addToLog(message=message, level=level, messageCode='arelle:pluginLoadingError')


def modulesWithNewerFileDates():
    names = set()
    for moduleName, moduleInfo in pluginConfig["modules"].items():
        freshenedFilename = _cntlr.webCache.getfilename(moduleInfo["moduleURL"], checkModifiedTime=True, normalize=True, base=_pluginBase)
        try:
            if os.path.isdir(freshenedFilename): # if freshenedFilename is a directory containing an __init__.py file, open that instead
                if os.path.isfile(os.path.join(freshenedFilename, "__init__.py")):
                    freshenedFilename = os.path.join(freshenedFilename, "__init__.py")
            elif not freshenedFilename.endswith(".py") and not os.path.exists(freshenedFilename) and os.path.exists(freshenedFilename + ".py"):
                freshenedFilename += ".py" # extension module without .py suffix
            if os.path.exists(freshenedFilename):
                if moduleInfo["fileDate"] < time.strftime('%Y-%m-%dT%H:%M:%S UTC', time.gmtime(os.path.getmtime(freshenedFilename))):
                    names.add(moduleInfo["name"])
            else:
                _msg = _("File not found for '{name}' plug-in when checking for updated module info. Path: '{path}'") \
                    .format(name=moduleName, path=freshenedFilename)
                logPluginTrace(_msg, logging.ERROR)
        except Exception as err:
            _msg = _("Exception at plug-in method modulesWithNewerFileDates: {error}").format(error=err)
            logPluginTrace(_msg, logging.ERROR)

    return names

def freshenModuleInfos():
    # for modules with different date-times, re-load module info
    missingEnabledModules = []
    for moduleName, moduleInfo in pluginConfig["modules"].items():
        moduleEnabled = moduleInfo["status"] == "enabled"
        freshenedFilename = _cntlr.webCache.getfilename(moduleInfo["moduleURL"], checkModifiedTime=True, normalize=True, base=_pluginBase)
        try: # check if moduleInfo cached may differ from referenced moduleInfo
            if os.path.isdir(freshenedFilename): # if freshenedFilename is a directory containing an __ini__.py file, open that instead
                if os.path.isfile(os.path.join(freshenedFilename, "__init__.py")):
                    freshenedFilename = os.path.join(freshenedFilename, "__init__.py")
            elif not freshenedFilename.endswith(".py") and not os.path.exists(freshenedFilename) and os.path.exists(freshenedFilename + ".py"):
                freshenedFilename += ".py" # extension module without .py suffix
            if os.path.exists(freshenedFilename):
                if moduleInfo["fileDate"] != time.strftime('%Y-%m-%dT%H:%M:%S UTC', time.gmtime(os.path.getmtime(freshenedFilename))):
                    freshenedModuleInfo = moduleModuleInfo(moduleURL=moduleInfo["moduleURL"], reload=True)
                    if freshenedModuleInfo is not None:
                        if freshenedModuleInfo["name"] == moduleName:
                            pluginConfig["modules"][moduleName] = freshenedModuleInfo
                        else:
                            # Module has been re-named
                            if moduleEnabled:
                                missingEnabledModules.append(moduleName)
            # User can avoid pruning by disabling plugin
            elif moduleEnabled:
                missingEnabledModules.append(moduleName)
            else:
                _msg = _("File not found for '{name}' plug-in when attempting to update module info. Path: '{path}'")\
                    .format(name=moduleName, path=freshenedFilename)
                logPluginTrace(_msg, logging.ERROR)
        except Exception as err:
            _msg = _("Exception at plug-in method freshenModuleInfos: {error}").format(error=err)
            logPluginTrace(_msg, logging.ERROR)
    for moduleName in missingEnabledModules:
        removePluginModule(moduleName)
        # Try re-adding plugin modules by name (for plugins that moved from built-in to pip installed)
        moduleInfo = addPluginModule(moduleName)
        if moduleInfo:
            pluginConfig["modules"][moduleInfo["name"]] = moduleInfo
            loadModule(moduleInfo)
            logPluginTrace(_("Reloaded plugin that failed loading: {} {}").format(moduleName, moduleInfo), logging.INFO)
        else:
            logPluginTrace(_("Removed plugin that failed loading (plugin may have been archived): {}").format(moduleName), logging.ERROR)
    save(_cntlr)


def normalizeModuleFilename(moduleFilename: str) -> str | None:
    """
    Attempts to find python script as plugin entry point.
    A value will be returned
      if `moduleFilename` exists as-is,
      if `moduleFilename` is a directory containing __init__.py, or
      if `moduleFilename` with .py extension added exists
    :param moduleFilename:
    :return: Normalized filename, if exists
    """
    if os.path.isfile(moduleFilename):
        # moduleFilename exists as-is, use it
        return moduleFilename
    if os.path.isdir(moduleFilename):
        # moduleFilename is a directory, only valid script is __init__.py contained inside
        initPath = os.path.join(moduleFilename, "__init__.py")
        if os.path.isfile(initPath):
            return initPath
        else:
            return None
    if not moduleFilename.endswith(".py"):
        # moduleFilename is not a file or directory, try adding .py
        pyPath = moduleFilename + ".py"
        if os.path.exists(pyPath):
            return pyPath
    return None


def getModuleFilename(moduleURL: str, reload: bool, normalize: bool, base: str | None) -> tuple[str | None, EntryPoint | None]:
    #TODO several directories, eg User Application Data
    moduleFilename = _cntlr.webCache.getfilename(moduleURL, reload=reload, normalize=normalize, base=base)
    if moduleFilename:
        # `moduleURL` was mapped to a local filepath
        moduleFilename = normalizeModuleFilename(moduleFilename)
        if moduleFilename:
            # `moduleFilename` normalized to an existing script
            return moduleFilename, None
    if base and not _isAbsoluteModuleURL(moduleURL):
        # Search for a matching plugin deeper in the plugin directory tree.
        # Handles cases where a plugin exists in a nested structure, such as
        # when a developer clones an entire repository into the plugin directory.
        # Example: arelle/plugin/xule/plugin/xule/__init__.py
        for path in glob("**/" + moduleURL.replace('\\', '/'), recursive=True):
            if normalizedPath := normalizeModuleFilename(path):
                return normalizedPath, None
    # `moduleFilename` did not map to a local filepath or did not normalize to a script
    # Try using `moduleURL` to search for pip-installed entry point
    entryPointRef = EntryPointRef.get(moduleURL)
    if entryPointRef is not None:
        return entryPointRef.moduleFilename, entryPointRef.entryPoint
    return None, None


def parsePluginInfo(moduleURL: str, moduleFilename: str, entryPoint: EntryPoint | None) -> dict | None:
    moduleDir, moduleName = os.path.split(moduleFilename)
    f = arelle.FileSource.openFileStream(_cntlr, moduleFilename)
    tree = ast.parse(f.read(), filename=moduleFilename)
    constantStrings = {}
    functionDefNames = set()
    methodDefNamesByClass = defaultdict(set)
    moduleImports = []
    moduleInfo = {"name":None}
    isPlugin = False
    for item in tree.body:
        if isinstance(item, ast.Assign):
            try:
                attr = item.targets[0].id
            except AttributeError:
                # Not plugininfo
                continue
            if attr == "__pluginInfo__":
                isPlugin = True
                f.close()
                classMethods = []
                importURLs = []
                for i, key in enumerate(item.value.keys):
                    _key = key.value
                    _value = item.value.values[i]
                    _valueType = _value.__class__.__name__
                    if _key == "import":
                        if _valueType == 'Constant':
                            importURLs.append(_value.value)
                        elif _valueType in ("List", "Tuple"):
                            for elt in _value.elts:
                                importURLs.append(elt.value)
                    elif _valueType == 'Constant':
                        moduleInfo[_key] = _value.value
                    elif _valueType == 'Name':
                        if _value.id in constantStrings:
                            moduleInfo[_key] = constantStrings[_value.id]
                        elif _value.id in functionDefNames:
                            classMethods.append(_key)
                    elif _valueType == 'Attribute':
                        if _value.attr in methodDefNamesByClass[_value.value.id]:
                            classMethods.append(_key)
                    elif _valueType in ("List", "Tuple"):
                        values = [elt.value for elt in _value.elts]
                        if _key == "imports":
                            importURLs = values
                        else:
                            moduleInfo[_key] = values

                moduleInfo['classMethods'] = classMethods
                moduleInfo['importURLs'] = importURLs
                moduleInfo["moduleURL"] = moduleURL
                moduleInfo["path"] = moduleFilename
                moduleInfo["status"] = 'enabled'
                moduleInfo["fileDate"] = time.strftime('%Y-%m-%dT%H:%M:%S UTC', time.gmtime(os.path.getmtime(moduleFilename)))
                if entryPoint:
                    moduleInfo["moduleURL"] = moduleFilename  # pip-installed plugins need absolute filepath
                    moduleInfo["entryPoint"] = {
                        "module": entryPoint.module,
                        "name": entryPoint.name,
                        "version": entryPoint.dist.version if hasattr(entryPoint, 'dist') else None,
                    }
                    if not moduleInfo.get("version"):
                        moduleInfo["version"] = entryPoint.dist.version  # If no explicit version, retrieve from entry point
            elif isinstance(item.value, ast.Constant) and isinstance(item.value.value, str):  # possible constant used in plugininfo, such as VERSION
                for assignmentName in item.targets:
                    constantStrings[assignmentName.id] = item.value.value
        elif isinstance(item, ast.ImportFrom):
            if item.level == 1: # starts with .
                if item.module is None:  # from . import module1, module2, ...
                    for importee in item.names:
                        if importee.name == '*': #import all submodules
                            for _file in os.listdir(moduleDir):
                                if _file != moduleName and os.path.isfile(_file) and _file.endswith(".py"):
                                    moduleImports.append(_file)
                        elif (os.path.isfile(os.path.join(moduleDir, importee.name + ".py"))
                              and importee.name not in moduleImports):
                            moduleImports.append(importee.name)
                else:
                    modulePkgs = item.module.split('.')
                    modulePath = os.path.join(*modulePkgs)
                    if (os.path.isfile(os.path.join(moduleDir, modulePath) + ".py")
                            and modulePath not in moduleImports):
                        moduleImports.append(modulePath)
                    for importee in item.names:
                        _importeePfxName = os.path.join(modulePath, importee.name)
                        if (os.path.isfile(os.path.join(moduleDir, _importeePfxName) + ".py")
                                and _importeePfxName not in moduleImports):
                            moduleImports.append(_importeePfxName)
        elif isinstance(item, ast.FunctionDef): # possible functionDef used in plugininfo
            functionDefNames.add(item.name)
        elif isinstance(item, ast.ClassDef):  # possible ClassDef used in plugininfo
            for classItem in item.body:
                if isinstance(classItem, ast.FunctionDef):
                    methodDefNamesByClass[item.name].add(classItem.name)
    moduleInfo["moduleImports"] = moduleImports
    f.close()
    return moduleInfo if isPlugin else None


def moduleModuleInfo(
        moduleURL: str | None = None,
        entryPoint: EntryPoint | None = None,
        reload: bool = False,
        parentImportsSubtree: bool = False) -> dict | None:
    """
    Generates a module info dict based on the provided `moduleURL` or `entryPoint`
    Exactly one of "moduleURL" or "entryPoint" must be provided, otherwise a RuntimeError will be thrown.

    When `moduleURL` is provided, it will be treated as a file path and will attempt to be normalized and
    mapped to an existing plugin based on file location. If `moduleURL` fails to be mapped to an existing
    plugin on its own, it will instead be used to search for an entry point. If found, this function will
    proceed as if that entry point was provided for `entryPoint`.

    When `entryPoint` is provided, it's location and other details will be used to generate the module info
    dictionary.

    :param moduleURL: A URL that loosely maps to the file location of a plugin (may be transformed)
    :param entryPoint: An `EntryPoint` instance
    :param reload:
    :param parentImportsSubtree:
    :return:s
    """
    if (moduleURL is None) == (entryPoint is None):
        raise RuntimeError('Exactly one of "moduleURL" or "entryPoint" must be provided')
    if entryPoint:
        # If entry point is provided, use it to retrieve `moduleFilename`
        moduleFilename = moduleURL = entryPoint.load()()
    else:
        # Otherwise, we will verify the path before continuing
        moduleFilename, entryPoint = getModuleFilename(moduleURL, reload=reload, normalize=True, base=_pluginBase)

    if moduleFilename:
        try:
            logPluginTrace(f"Scanning module for plug-in info: {moduleFilename}", logging.INFO)
            moduleInfo = parsePluginInfo(moduleURL, moduleFilename, entryPoint)
            if moduleInfo is None:
                return None

            moduleDir, moduleName = os.path.split(moduleFilename)
            importURLs = moduleInfo["importURLs"]
            del moduleInfo["importURLs"]
            moduleImports = moduleInfo["moduleImports"]
            del moduleInfo["moduleImports"]
            moduleImportsSubtree = False
            mergedImportURLs = []

            for url in importURLs:
                if url.startswith("module_import"):
                    for moduleImport in moduleImports:
                        mergedImportURLs.append(moduleImport + ".py")
                    if url == "module_import_subtree":
                        moduleImportsSubtree = True
                elif url == "module_subtree":
                    for _dir in os.listdir(moduleDir):
                        subtreeModule = os.path.join(moduleDir,_dir)
                        if os.path.isdir(subtreeModule) and _dir != "__pycache__":
                            mergedImportURLs.append(subtreeModule)
                else:
                    mergedImportURLs.append(url)
            if parentImportsSubtree and not moduleImportsSubtree:
                moduleImportsSubtree = True
                for moduleImport in moduleImports:
                    mergedImportURLs.append(moduleImport + ".py")
            imports = []
            for url in mergedImportURLs:
                importURL = url
                if not _isAbsoluteModuleURL(url):
                    # Handle relative imports when plugin is loaded from external directory.
                    # When EDGAR/render imports EDGAR/validate, this works if EDGAR is in the plugin directory
                    # but fails if loaded externally (e.g., dev repo clone at /dev/path/to/EDGAR/).
                    # Solution: Find common path segments to resolve /dev/path/to/EDGAR/validate
                    # from the importing module at /dev/path/to/EDGAR/render.
                    modulePath = Path(moduleFilename)
                    importPath = Path(url)
                    if importPath.parts:
                        importFirstPart = importPath.parts[0]
                        for i, modulePathPart in enumerate(reversed(modulePath.parts)):
                            if modulePathPart != importFirstPart:
                                continue
                            # Found a potential branching point, construct and check a new path
                            candidateImportURL = str(modulePath.parents[i] / importPath)
                            if normalizeModuleFilename(candidateImportURL):
                                importURL = candidateImportURL
                importModuleInfo = moduleModuleInfo(moduleURL=importURL, reload=reload, parentImportsSubtree=moduleImportsSubtree)
                if importModuleInfo:
                    importModuleInfo["isImported"] = True
                    imports.append(importModuleInfo)
            moduleInfo["imports"] = imports
            logPluginTrace(f"Successful module plug-in info: {moduleFilename}", logging.INFO)
            return moduleInfo
        except Exception as err:
            _msg = _("Exception obtaining plug-in module info: {moduleFilename}\n{error}\n{traceback}").format(
                    error=err, moduleFilename=moduleFilename, traceback=traceback.format_tb(sys.exc_info()[2]))
            logPluginTrace(_msg, logging.ERROR)
    return None


def moduleInfo(pluginInfo):
    moduleInfo = {}
    for name, value in pluginInfo.items():
        if isinstance(value, str):
            moduleInfo[name] = value
        elif isinstance(value, types.FunctionType):
            moduleInfo.getdefault('classes', []).append(name)


def _isAbsoluteModuleURL(moduleURL: str) -> bool:
    return isAbsolute(moduleURL) or isLegacyAbs(moduleURL)


def _get_name_dir_prefix(modulePath: Path, packagePrefix: str = "") -> tuple[str, str, str] | tuple[None, None, None]:
    """Get the name, directory and prefix of a module."""
    moduleName = None
    moduleDir = None
    packageImportPrefix = None
    initFileName = "__init__.py"

    if modulePath.is_file() and modulePath.name == initFileName:
        modulePath = modulePath.parent

    if modulePath.is_dir() and (modulePath / initFileName).is_file():
        moduleName = modulePath.name
        moduleDir = str(modulePath.parent)
        packageImportPrefix = moduleName + "."
    elif modulePath.is_file() and modulePath.suffix == ".py":
        moduleName = modulePath.stem
        moduleDir = str(modulePath.parent)
        packageImportPrefix = packagePrefix

    return (moduleName, moduleDir, packageImportPrefix)

def _get_location(moduleDir: str, moduleName: str) -> str:
    """Get the file name of a plugin."""
    module_name_path = Path(f"{moduleDir}/{moduleName}.py")
    if os.path.isfile(module_name_path):
        return module_name_path

    return Path(f"{moduleDir}/{moduleName}/__init__.py")

def _find_and_load_module(moduleDir: str, moduleName: str) -> ModuleType | None:
    """Load a module based on name and directory."""
    location = _get_location(moduleDir=moduleDir, moduleName=moduleName)
    spec = importlib.util.spec_from_file_location(name=moduleName, location=location)

    # spec_from_file_location returns ModuleSpec or None.
    # spec.loader returns Loader or None.
    # We want to make sure neither of them are None before proceeding
    if spec is None or spec.loader is None:
        raise ModuleNotFoundError("Unable to load module")

    module = importlib.util.module_from_spec(spec)
    sys.modules[moduleName] = module # This line is required before exec_module
    spec.loader.exec_module(sys.modules[moduleName])

    return sys.modules[moduleName]

def loadModule(moduleInfo: dict[str, Any], packagePrefix: str="") -> None:
    name = moduleInfo['name']
    moduleURL = moduleInfo['moduleURL']
    modulePath = Path(moduleInfo['path'])

    moduleName, moduleDir, packageImportPrefix = _get_name_dir_prefix(modulePath, packagePrefix)

    if all(p is None for p in [moduleName, moduleDir, packageImportPrefix]):
        _cntlr.addToLog(message=_ERROR_MESSAGE_IMPORT_TEMPLATE.format(name), level=logging.ERROR)
    else:
        try:
            module = _find_and_load_module(moduleDir=moduleDir, moduleName=moduleName)
            pluginInfo = module.__pluginInfo__.copy()
            elementSubstitutionClasses = None
            if name == pluginInfo.get('name'):
                pluginInfo["moduleURL"] = moduleURL
                modulePluginInfos[name] = pluginInfo
                if 'localeURL' in pluginInfo:
                    # set L10N internationalization in loaded module
                    localeDir = os.path.dirname(module.__file__) + os.sep + pluginInfo['localeURL']
                    try:
                        _gettext = gettext.translation(pluginInfo['localeDomain'], localeDir, getLanguageCodes())
                    except OSError:
                        def _gettext(x):
                            return x # no translation
                else:
                    def _gettext(x):
                        return x
                for key, value in pluginInfo.items():
                    if key == 'name':
                        if name:
                            pluginConfig['modules'][name] = moduleInfo
                    elif isinstance(value, types.FunctionType):
                        classModuleNames = pluginConfig['classes'].setdefault(key, [])
                        if name and name not in classModuleNames:
                            classModuleNames.append(name)
                    if key == 'ModelObjectFactory.ElementSubstitutionClasses':
                        elementSubstitutionClasses = value
                module._ = _gettext
                global pluginConfigChanged
                pluginConfigChanged = True
            if elementSubstitutionClasses:
                try:
                    from arelle.ModelObjectFactory import elementSubstitutionModelClass
                    elementSubstitutionModelClass.update(elementSubstitutionClasses)
                except Exception as err:
                    _msg = _("Exception loading plug-in {name}: processing ModelObjectFactory.ElementSubstitutionClasses").format(
                            name=name, error=err)
                    logPluginTrace(_msg, logging.ERROR)
            for importModuleInfo in moduleInfo.get('imports', EMPTYLIST):
                loadModule(importModuleInfo, packageImportPrefix)
        except (AttributeError, ImportError, FileNotFoundError, ModuleNotFoundError, TypeError, SystemError) as err:
            # Send a summary of the error to the logger and retain the stacktrace for stderr
            _cntlr.addToLog(message=_ERROR_MESSAGE_IMPORT_TEMPLATE.format(name), level=logging.ERROR)

            _msg = _("Exception loading plug-in {name}: {error}\n{traceback}").format(
                    name=name, error=err, traceback=traceback.format_tb(sys.exc_info()[2]))
            logPluginTrace(_msg, logging.ERROR)


def hasPluginWithHook(name: str) -> bool:
    return next(pluginClassMethods(name), None) is not None


def pluginClassMethods(className: str) -> Iterator[Callable[..., Any]]:
    if pluginConfig:
        try:
            pluginMethodsForClass = pluginMethodsForClasses[className]
        except KeyError:
            # load all modules for class
            pluginMethodsForClass = []
            modulesNamesLoaded = set()
            if className in pluginConfig["classes"]:
                for moduleName in pluginConfig["classes"].get(className):
                    if moduleName and moduleName in pluginConfig["modules"] and moduleName not in modulesNamesLoaded:
                        modulesNamesLoaded.add(moduleName) # prevent multiply executing same class
                        moduleInfo = pluginConfig["modules"][moduleName]
                        if moduleInfo["status"] == "enabled":
                            if moduleName not in modulePluginInfos:
                                loadModule(moduleInfo)
                            if moduleName in modulePluginInfos:
                                pluginInfo = modulePluginInfos[moduleName]
                                if className in pluginInfo:
                                    pluginMethodsForClass.append(pluginInfo[className])
            pluginMethodsForClasses[className] = pluginMethodsForClass
        yield from pluginMethodsForClass


def addPluginModule(name: str) -> dict[str, Any] | None:
    """
    Discover plugin entry points with given name.
    :param name: The name to search for
    :return: The module information dictionary, if added. Otherwise, None.
    """
    entryPointRef = EntryPointRef.get(name)
    pluginModuleInfo = None
    if entryPointRef:
        pluginModuleInfo = entryPointRef.createModuleInfo()
    if not pluginModuleInfo or not pluginModuleInfo.get("name"):
        pluginModuleInfo = moduleModuleInfo(moduleURL=name)
    return addPluginModuleInfo(pluginModuleInfo)


def reloadPluginModule(name):
    if name in pluginConfig["modules"]:
        url = pluginConfig["modules"][name].get("moduleURL")
        if url:
            moduleInfo = moduleModuleInfo(moduleURL=url, reload=True)
            if moduleInfo:
                addPluginModule(url)
                return True
    return False

def removePluginModule(name):
    moduleInfo = pluginConfig["modules"].get(name)
    if moduleInfo and name:
        def _removePluginModule(moduleInfo):
            _name = moduleInfo.get("name")
            if _name:
                for classMethod in moduleInfo["classMethods"]:
                    classMethods = pluginConfig["classes"].get(classMethod)
                    if classMethods and _name and _name in classMethods:
                        classMethods.remove(_name)
                        if not classMethods: # list has become unused
                            del pluginConfig["classes"][classMethod] # remove class
                for importModuleInfo in moduleInfo.get('imports', EMPTYLIST):
                    _removePluginModule(importModuleInfo)
                pluginConfig["modules"].pop(_name, None)
        _removePluginModule(moduleInfo)
        global pluginConfigChanged
        pluginConfigChanged = True
        return True
    return False # unable to remove


def addPluginModuleInfo(plugin_module_info: dict[str, Any]) -> dict[str, Any] | None:
    """
    Given a dictionary containing module information, loads plugin info into `pluginConfig`
    :param plugin_module_info: Dictionary of module info fields. See comment block in PluginManager.py for structure.
    :return: The module information dictionary, if added. Otherwise, None.
    """
    if not plugin_module_info or not plugin_module_info.get("name"):
        return None
    name = plugin_module_info["name"]
    removePluginModule(name)  # remove any prior entry for this module

    def _addPluginSubModule(subModuleInfo: dict[str, Any]):
        """
        Inline function for recursively exploring module imports
        :param subModuleInfo: Module information to add.
        :return:
        """
        _name = subModuleInfo.get("name")
        if not _name:
            return
        # add classes
        for classMethod in subModuleInfo["classMethods"]:
            classMethods = pluginConfig["classes"].setdefault(classMethod, [])
            _name = subModuleInfo["name"]
            if _name and _name not in classMethods:
                classMethods.append(_name)
        for importModuleInfo in subModuleInfo.get('imports', EMPTYLIST):
            _addPluginSubModule(importModuleInfo)
        pluginConfig["modules"][_name] = subModuleInfo

    _addPluginSubModule(plugin_module_info)
    global pluginConfigChanged
    pluginConfigChanged = True
    return plugin_module_info


_entryPointRefCache: list[EntryPointRef] | None = None
_entryPointRefAliasCache: dict[str, list[EntryPointRef]] | None = None
_entryPointRefSearchTermEndings = [
    '/__init__.py',
    '.py'
    '/'
]


@dataclass
class EntryPointRef:
    aliases: set[str]
    entryPoint: EntryPoint | None
    moduleFilename: str | None
    moduleInfo: dict | None

    def createModuleInfo(self) -> dict | None:
        """
        Creates a module information dictionary from the entry point ref.
        :return: A module inforomation dictionary
        """
        if self.entryPoint is not None:
            return moduleModuleInfo(entryPoint=self.entryPoint)
        return moduleModuleInfo(moduleURL=self.moduleFilename)

    @staticmethod
    def fromEntryPoint(entryPoint: EntryPoint) -> EntryPointRef | None:
        """
        Given an entry point, retrieves the subset of information from __pluginInfo__ necessary to
        determine if the entry point should be imported as a plugin.
        :param entryPoint:
        :return:
        """
        pluginUrlFunc = entryPoint.load()
        pluginUrl = pluginUrlFunc()
        return EntryPointRef.fromFilepath(pluginUrl, entryPoint)

    @staticmethod
    def fromFilepath(filepath: str, entryPoint: EntryPoint | None = None) -> EntryPointRef | None:
        """
        Given a filepath, retrieves a subset of information from __pluginInfo__ necessary to
        determine if the entry point should be imported as a plugin.
        :param filepath: Path to plugin, can be a directory or .py filepath
        :param entryPoint: Optional entry point information to include in aliases/moduleInfo
        :return:
        """
        moduleFilename = _cntlr.webCache.getfilename(filepath)
        if moduleFilename:
            moduleFilename = normalizeModuleFilename(moduleFilename)
        aliases = set()
        if entryPoint:
            aliases.add(entryPoint.name)
        moduleInfo: dict | None = None
        if moduleFilename:
            moduleInfo = parsePluginInfo(moduleFilename, moduleFilename, entryPoint)
            if moduleInfo is None:
                return None
            if "name" in moduleInfo:
                aliases.add(moduleInfo["name"])
            if "aliases" in moduleInfo:
                aliases |= set(moduleInfo["aliases"])
        return EntryPointRef(
            aliases={EntryPointRef._normalizePluginSearchTerm(a) for a in aliases},
            entryPoint=entryPoint,
            moduleFilename=moduleFilename,
            moduleInfo=moduleInfo,
        )

    @staticmethod
    def discoverAll() -> list[EntryPointRef]:
        """
        Retrieve all plugin entry points, cached on first run.
        :return: List of all discovered entry points.
        """
        global _entryPointRefCache
        if _entryPointRefCache is None:
            assert _pluginBase is not None
            _entryPointRefCache = EntryPointRef._discoverBuiltIn([], cast(str, _pluginBase)) + EntryPointRef._discoverInstalled()
        return _entryPointRefCache

    @staticmethod
    def _discoverBuiltIn(entryPointRefs: list[EntryPointRef], directory: str) -> list[EntryPointRef]:
        """
        Recursively retrieve all plugin entry points in the given directory.
        :param entryPointRefs: Working list of entry point refs to append to.
        :param directory: Directory to search for entry points within.
        :return: List of discovered entry points.
        """
        for fileName in sorted(os.listdir(directory)):
            if fileName in (".", "..", "__pycache__", "__init__.py", ".DS_Store", "site-packages"):
                continue  # Ignore these entries
            filePath = os.path.join(directory, fileName)
            if os.path.isdir(filePath):
                EntryPointRef._discoverBuiltIn(entryPointRefs, filePath)
            if os.path.isfile(filePath) and os.path.basename(filePath).endswith(".py"):
                # If `filePath` references .py file directly, use it
                moduleFilePath = filePath
            elif os.path.isdir(filePath) and os.path.exists(initFilePath := os.path.join(filePath, "__init__.py")):
                # Otherwise, if `filePath` is a directory containing `__init__.py`, use that
                moduleFilePath = initFilePath
            else:
                continue
            entryPointRef = EntryPointRef.fromFilepath(moduleFilePath)
            if entryPointRef is not None:
                entryPointRefs.append(entryPointRef)
        return entryPointRefs

    @staticmethod
    def _discoverInstalled() -> list[EntryPointRef]:
        """
        Retrieve all installed plugin entry points.
        :return: List of all discovered entry points.
        """
        entryPoints: list[EntryPoint]
        if sys.version_info < (3, 10):
            entryPoints = [e for e in entry_points().get('arelle.plugin', [])]
        else:
            entryPoints = list(entry_points(group='arelle.plugin'))
        entryPointRefs = []
        for entryPoint in entryPoints:
            entryPointRef = EntryPointRef.fromEntryPoint(entryPoint)
            if entryPointRef is not None:
                entryPointRefs.append(entryPointRef)
        return entryPointRefs

    @staticmethod
    def get(search: str) -> EntryPointRef | None:
        """
        Retrieve an entry point ref with a matching name or alias.
        May return None of no matches are found.
        Throws an exception if multiple entry point refs match the search term.
        :param search: Only retrieve entry point matching the given search text.
        :return: Matching entry point ref, if found.
        """
        entryPointRefs = EntryPointRef.search(search)
        if len(entryPointRefs) == 0:
            return None
        elif len(entryPointRefs) > 1:
            paths = [r.moduleFilename for r in entryPointRefs]
            raise Exception(_('Multiple entry points matched search term "{}": {}').format(search, paths))
        return entryPointRefs[0]

    @staticmethod
    def _normalizePluginSearchTerm(search: str) -> str:
        """
        Normalizes the given search term or searchable text by:
          Making slashes consistent
          Removing common endings
        :param search: Search term or searchable text
        :return: Normalized string
        """
        search = search.replace('\\', '/')
        while True:
            for ending in _entryPointRefSearchTermEndings:
                if search.endswith(ending):
                    search = search[:-len(ending)]
                    break
            return search.lower()

    @staticmethod
    def search(search: str) -> list[EntryPointRef] | None:
        """
        Retrieve entry point module information matching provided search text.
        A map of aliases to matching entry points is cached on the first run.
        :param search: Only retrieve entry points matching the given search text.
        :return: List of matching module infos.
        """
        global _entryPointRefAliasCache
        if _entryPointRefAliasCache is None:
            entryPointRefAliasCache = defaultdict(list)
            entryPointRefs = EntryPointRef.discoverAll()
            for entryPointRef in entryPointRefs:
                for alias in entryPointRef.aliases:
                    entryPointRefAliasCache[alias].append(entryPointRef)
            _entryPointRefAliasCache = entryPointRefAliasCache
        search = EntryPointRef._normalizePluginSearchTerm(search)
        return _entryPointRefAliasCache.get(search, [])



"""
Save HTML EBA Tables is an example of a plug-in to both GUI menu and command line/web service
that will save a directory containing HTML Tablesets with an EBA index page.

See COPYRIGHT.md for copyright information.
"""

import os
import threading
from operator import itemgetter
from optparse import OptionParser
from tkinter import Menu
from typing import Any

from lxml import etree

from arelle import Version, XbrlConst, XmlUtil
from arelle.CntlrCmdLine import CntlrCmdLine
from arelle.CntlrWinMain import CntlrWinMain
from arelle.ModelDocument import Type
from arelle.ModelObjectFactory import parser
from arelle.ModelRenderingObject import DefnMdlTable
from arelle.ModelXbrl import ModelXbrl
from arelle.rendering import RenderingEvaluator
from arelle.RuntimeOptions import RuntimeOptions
from arelle.typing import TypeGetText
from arelle.utils.PluginHooks import PluginHooks
from arelle.ViewFileRenderedGrid import viewRenderedGrid

_: TypeGetText

MENU_HTML = """<!DOCTYPE html>
<html>
<head>
  <style>
    body {
      margin: 0;
      padding: 10px;
      font-family: Arial, sans-serif;
      color: #243e5e;
    }
    .nav-list-menu-ul {
      list-style-type: none;
      padding: 0;
      margin: 0;
    }
    .nav-list-menu-li {
      margin: 5px 0;
      padding: 5px;
      border-bottom: 1px solid #eee;
    }
    .nav-list-menu-link {
      text-decoration: none;
      cursor: pointer;
      display: block;
      background: none;
      border: none;
      padding: 0;
    }
    .nav-list-menu-link:hover {
      text-decoration: underline;
    }
  </style>
</head>
<body>
    <ul class="nav-list-menu-ul"/>
</body>
</html>
"""

CENTER_LANDING_HTML = """<!DOCTYPE html>
<html>
<head>
  <style>
    body {
      margin: 0;
      padding: 20px;
      font-family: Arial, sans-serif;
    }
    #page-title {
      margin-bottom: 20px;
    }
    #page-title h1 {
      color: #243e5e;
      margin-top: 0;
    }
    #content-center {
      margin-top: 20px;
      line-height: 1.5;
    }
  </style>
</head>
<body>
  <div id="page-title">
    <h1>Taxonomy Tables Viewer</h1>
  </div>
  <div id="content-center">
    <p>Please select tables to view by clicking in the left column.</p>
  </div>
</body>
</html>
"""

TABLE_CSS_EXTRAS = """
table {background:#fff}
"""


def indexFileHTML(indexBaseName: str) -> str:
    return f'''<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>EBA - Tablesets</title>
  <style>
    html, body {{
      margin: 0;
      padding: 0;
      height: 100%;
      width: 100%;
      font-family: Arial, sans-serif;
      display: flex;
      flex-direction: column;
    }}
    #header {{
      background: rgb(36, 62, 94);
      color: rgb(255, 255, 255);
      height: 40px;
    }}
    #header h1 {{
      font-size: 1.5em;
      margin: 0.25em;
    }}
    #main-container {{
      display: flex;
      flex: 1;
      height: calc(100vh - 40px);
    }}
    #menu-container {{
      width: 360px;
      border-right: 2px solid #243e5e;
      overflow-y: auto;
      box-sizing: border-box;
    }}
    #content-container {{
      flex: 1;
      overflow: auto;
      box-sizing: border-box;
    }}
    iframe {{
      border: none;
      width: 100%;
      height: 100%;
    }}
  </style>
  <script>
    function loadContent(url) {{
      document.getElementById('content-frame').src = url;
    }}
  </script>
</head>
<body>
  <div id="header"><h1>EBA - Tablesets</h1></div>
  <div id="main-container">
    <div id="menu-container">
      <iframe src="{indexBaseName}FormsFrame.html" width="100%" height="100%" frameborder="0" id="menu-frame"></iframe>
    </div>
    <div id="content-container">
      <iframe src="{indexBaseName}CenterLanding.html" width="100%" height="100%" frameborder="0" id="content-frame"></iframe>
    </div>
  </div>
</body>
</html>
'''


def generateHtmlEbaTablesetFiles(dts: ModelXbrl, indexFile: str, lang: str = "en") -> None:
    try:
        numTableFiles = 0
        _parser = parser(dts, None)[0]
        menuFrameDocument = etree.fromstring(MENU_HTML, parser=_parser, base_url=indexFile)
        listElt = menuFrameDocument.find(".//ul")
        assert listElt is not None, "No list element in index document"

        indexBase = indexFile.rpartition(".")[0]
        groupTableRels = dts.modelXbrl.relationshipSet(XbrlConst.euGroupTable)
        modelTables = []

        def viewTable(modelTable: DefnMdlTable) -> None:
            if modelTable is None:
                return
            tableId = modelTable.id or ""
            if isinstance(modelTable, DefnMdlTable):
                dts.modelManager.cntlr.addToLog("viewing: " + tableId)
                tblFile = os.path.join(os.path.dirname(indexFile), tableId + ".html")
                tableName = tableId
                if tableName.startswith("eba_t"):
                    tableName = tableName.removeprefix("eba_t")
                elif tableName.startswith("srb_t"):
                    tableName = tableName.removeprefix("srb_t")
                viewRenderedGrid(dts, tblFile, lang=lang, cssExtras=TABLE_CSS_EXTRAS, table=tableName)  # type: ignore[no-untyped-call]

                elt = etree.SubElement(listElt, "li")
                elt.set("class", "nav-list-menu-li")
                elt.set("id", tableId)
                elt = etree.SubElement(elt, "button")
                elt.text = modelTable.genLabel(lang=lang, strip=True)
                elt.set("class", "nav-list-menu-link")
                elt.set("onClick", f"javascript:parent.loadContent('{tableId}.html');")
                elt.text = modelTable.genLabel(lang=lang, strip=True)
            else:
                elt = etree.SubElement(listElt, "li")
                elt.set("id", tableId)
                elt.text = modelTable.label(lang=lang, strip=True)

            for rel in groupTableRels.fromModelObject(modelTable):
                viewTable(rel.toModelObject)

        for rootConcept in groupTableRels.rootConcepts:
            sourceline = 0
            for rel in dts.modelXbrl.relationshipSet(XbrlConst.euGroupTable).fromModelObject(rootConcept):
                sourceline = rel.sourceline
                break
            modelTables.append((rootConcept, sourceline))

        for modelTable, _order in sorted(modelTables, key=itemgetter(1)):
            viewTable(modelTable)

        with open(indexBase + "FormsFrame.html", "w", encoding="utf-8") as fh:
            XmlUtil.writexml(fh, menuFrameDocument, encoding="utf-8")

        with open(indexFile, "w", encoding="utf-8") as fh:
            fh.write(indexFileHTML(os.path.basename(indexBase)))

        with open(indexBase + "CenterLanding.html", "w", encoding="utf-8") as fh:
            fh.write(CENTER_LANDING_HTML)

        dts.info(
            "info:saveEBAtables",
            _("Tables index file of %(entryFile)s has %(numberTableFiles)s table files with index file %(indexFile)s."),
            modelObject=dts,
            entryFile=dts.uri,
            numberTableFiles=numTableFiles,
            indexFile=indexFile,
        )

        dts.modelManager.showStatus(_("Saved EBA HTML Table Files"), 5000)
    except Exception as ex:
        dts.error(
            "exception",
            _("HTML EBA Tableset files generation exception: %(error)s"),
            error=ex,
            modelXbrl=dts,
            exc_info=True,
        )


def saveHtmlEbaTablesMenuCommand(cntlr: CntlrWinMain) -> None:
    if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
        cntlr.addToLog("No DTS loaded.")  # type: ignore[no-untyped-call]
        return

    assert cntlr.config is not None
    indexFile = cntlr.uiFileDialog(  # type: ignore[no-untyped-call]
        "save",
        title=_("arelle - Save HTML EBA Tables Index file"),
        initialdir=cntlr.config.setdefault("htmlEbaTablesFileDir", "."),
        filetypes=[(_("HTML index file .html"), "*.html")],
        defaultextension=".html",
    )
    if not isinstance(indexFile, str):
        return
    cntlr.config["htmlEbaTablesFileDir"] = os.path.dirname(indexFile)
    cntlr.saveConfig()

    thread = threading.Thread(
        target=lambda _dts=cntlr.modelManager.modelXbrl, _indexFile=indexFile: generateHtmlEbaTablesetFiles(
            _dts, _indexFile
        )
    )
    thread.daemon = True
    thread.start()


class SaveHtmlEbaTablesPlugin(PluginHooks):
    @staticmethod
    def cntlrWinMainMenuTools(
        cntlr: CntlrWinMain,
        menu: Menu,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        menu.add_command(label="Save HTML EBA Tables", underline=0, command=lambda: saveHtmlEbaTablesMenuCommand(cntlr))

    @staticmethod
    def cntlrCmdLineOptions(
        parser: OptionParser,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        parser.add_option(
            "--save-EBA-tablesets",
            action="store",
            dest="ebaTablesetIndexFile",
            help=_("Save HTML EBA Tablesets index file with provided filename."),
        )

    @staticmethod
    def cntlrCmdLineXbrlLoaded(
        cntlr: CntlrCmdLine,
        options: RuntimeOptions,
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        ebaTablesetIndexFile = getattr(options, "ebaTablesetIndexFile", None)
        modelDocType = getattr(modelXbrl.modelDocument, "type", None)
        if ebaTablesetIndexFile == "generateEBAFiles" and modelDocType in (Type.TESTCASESINDEX, Type.TESTCASE):
            cntlr.modelManager.generateEBAFiles = True  # type: ignore[attr-defined]

    @staticmethod
    def cntlrCmdLineXbrlRun(
        cntlr: CntlrCmdLine,
        options: RuntimeOptions,
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        ebaTablesetIndexFile = getattr(options, "ebaTablesetIndexFile", None)
        if ebaTablesetIndexFile is None or ebaTablesetIndexFile == "generateEBAFiles":
            return
        if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
            cntlr.addToLog("No taxonomy loaded.")
            return
        RenderingEvaluator.init(modelXbrl)  # type: ignore[no-untyped-call]
        generateHtmlEbaTablesetFiles(cntlr.modelManager.modelXbrl, ebaTablesetIndexFile)


__pluginInfo__ = {
    "name": "Save HTML EBA Tables",
    "version": "0.10",
    "description": "This plug-in adds a feature to a directory containing HTML Tablesets with an EBA index page.",
    "license": "Apache-2",
    "author": Version.authorLabel,
    "copyright": Version.copyrightLabel,
    # classes of mount points (required)
    "CntlrWinMain.Menu.Tools": SaveHtmlEbaTablesPlugin.cntlrWinMainMenuTools,
    "CntlrCmdLine.Options": SaveHtmlEbaTablesPlugin.cntlrCmdLineOptions,
    "CntlrCmdLine.Xbrl.Loaded": SaveHtmlEbaTablesPlugin.cntlrCmdLineXbrlLoaded,
    "CntlrCmdLine.Xbrl.Run": SaveHtmlEbaTablesPlugin.cntlrCmdLineXbrlRun,
}



"""
See COPYRIGHT.md for copyright information.

Implementation of ISO 17442:2012(E) Appendix A

"""

import string
from enum import Enum, auto

import regex as re


class LEIValidationResult(Enum):
    VALID = auto()
    INVALID_LEXICAL = auto()
    INVALID_CHECKSUM = auto()

    def description(self) -> str:
        return self.name.lower().replace("_", " ")


LEI_VALID = LEIValidationResult.VALID
LEI_INVALID_LEXICAL = LEIValidationResult.INVALID_LEXICAL
LEI_INVALID_CHECKSUM = LEIValidationResult.INVALID_CHECKSUM

_leiLexicalPattern = re.compile(r"[0-9A-Z]{18}[0-9]{2}", re.ASCII)
_requiredLEILength = 20

_leiToDigitTable = str.maketrans(
    dict(
        (ord(char), str(index))
        for index, char in enumerate(string.ascii_uppercase, start=len(string.digits))
    )
)

# The pattern comes from Solvency II documentation here https://dev.eiopa.europa.eu/Taxonomy/Full/2.3.0/PF/EIOPA_PensionFunds_Validations_2.3.0_PWD2.xlsx
_validLeiDespiteChecksumFailPatternString = r"^(029200720E3M3A4D6D01|029200758D5M0AI3F601|315700X8JQ3IU0NGK501|3157007SCCESQAUH5Z01|315700TCC9NTEP7J8Z01|13250000000000000301|315700ADOXDR5PCY5400|315700A0UB9Q7DOQIZ00|315700BM9Z39TNTGQW00|315700BZ5F7DRYG2UM00|315700B8401GTFFY6X01|315700DJ07P6OX10FK01|315700DKCD4QSKLAMO01|315700D23JL5C1DZNT00|315700EIYO2TLSEGQ700|315700ET7M7VQ4C84R00|315700EZSEA51937KX01|315700GXBGM8DBYKHF01|315700GZA843JXKTJ400|315700G5G24XYL1TXH00|315700HS7WJ1B0SUWM01|315700HXOEOK58E58P01|315700HZU4SMI8LZTU00|315700I3W2AFHP8MNQ01|315700JICZ3SY5SAXX00|315700JKZH6I0067ND01|315700JO5E28SRE00Q01|315700JXUHL9H2C3P700|315700LDDN3RM7Y2MP00|315700LWYOZNQ7V1T100|315700MBYPT6PGKO7M01|315700MW2F0KFR45QW01|315700M5843O6DU83901|315700NNMGS8F3P2CN00|315700N0VEIBHP0NPQ01|315700OASRCM664PAW01|315700OFW4YCOBNX4U01|315700OVW93X0T3HP200|315700O666JVNCQU9X00|315700PLI0I7W8IOV400|315700PN3J57ZUNF1V00|315700PXKOSX7WQV4N00|315700P4N9VSLK5QZV01|315700P40OV6BT045900|315700P6TZOLP92KN801|315700P89WR82VNB8Z00|315700QGM4XWZE1I5N01|315700Q1S8O1UORF9700|315700RK8M4FAHMYAP01|315700RTEHY362KXWJ00|315700SNUXK41WMW5J00|315700S22RGYRIEEOT00|315700S3TF79ALV82F01|315700TF5Z7T28HZJK01|315700TWGZ89LLSRS000|315700TXNX10N8XH4K00|315700T2EEQAPBO0C301|315700T6T49EDM16YO01|315700UJ6N4LGKLNPB00|315700UKZXWXEO126601|315700UYFD5GF9R13F01|315700VG7PTE9EJJRX01|315700VITYR7AL4M9S01|315700VMAJZ9JZTXNQ00|315700WH3YMKHCVYW201|315700WKCDF4QGRRO200|315700WR4IHOO1M5LP00|315700WYOZ6994UATN00|315700WZPEIS41QDKE00|315700XEFYMA5EZ0P500|315700XE21UYOA3GAC01|315700XI4Z8GF5BDUJ01|315700XSCP1S8WOD8E01|315700X40GNCOUWJYR00|315700YS6RQ5TF3VBP01|315700Y1W7W1JHAUBW00|315700Y5JNQMMUF5ID01|3157000VAJWZ3P8ZED00|3157001KM8GOU7PXZY01|3157001K2LAL04D87901|3157001MLDD3SDFQA901|3157001TPR6K4GBTLN00|31570010000000006400|31570010000000009601|31570010000000019301|31570010000000025800|31570010000000029001|31570010000000035500|31570010000000038701|31570010000000045200|31570010000000048401|31570010000000054900|31570010000000064600|31570010000000067801|31570010000000077501|31570010000000084000|31570010000000087201|31570010000000096901|31570010000000103400|31570010000000106601|31570010000000116301|31570010000000122800|31570010000000126001|3157002CKDQOCIHE5H01|31570020000000005900|31570029808HJVCNFA01|3157003FQSSGS9OZ9E01|3157004PTVTDOKB46401|3157004R6CH6C1P4KX00|3157005RUI28M8FANK00|3157005VJE7A3MBUS201|3157005WT1SENAE17R00|31570058O0Z320C4GZ00|3157006B6JVZ5DFMSN00|3157006DE3SPNIUY9K01|3157006FR3JBBOLOMX01|3157006I3B6RSTPQLI00|3157006KT1EZ15OIXW00|315700659AALVVLVIO01|31570067WSCDST0S3F01|3157008VKYORMUNC5O01|3157008ZY7CW6LVU5J00|3157009FTHFDDK7FHW01|3157009OVCV07O4HXM00|315700T8U7IU4W8J3A01|315700BBRQHDWX6SHZ00|3157008KD17KROO7UT01)$"

# Turn the _validLeiDespiteChecksumFailPattern in to an immutable set as that is much faster to check
_validLeiDespiteChecksumFailSet = frozenset(
    _validLeiDespiteChecksumFailPatternString[2:-2].split("|")
)


def checkLei(lei: str) -> LEIValidationResult:
    if len(lei) != _requiredLEILength:
        return LEIValidationResult.INVALID_LEXICAL

    if not _leiLexicalPattern.fullmatch(lei):
        return LEIValidationResult.INVALID_LEXICAL

    result = LEIValidationResult.INVALID_CHECKSUM

    if int(lei.translate(_leiToDigitTable)) % 97 == 1:
        result = LEIValidationResult.VALID

    if (
        result is LEIValidationResult.INVALID_CHECKSUM
        and lei in _validLeiDespiteChecksumFailSet
    ):
        result = LEIValidationResult.VALID

    return result


if __name__ == "__main__":
    import sys

    leis = sys.argv[1:]
    if not leis:
        raise SystemExit("Specify the LEI(s) you want to check as arguments.")

    for arg in sys.argv[1:]:
        print(f"{arg:20s} : {checkLei(arg).description()}")
